
import platform
from conan_unified_api.unified_api import ConanBaseUnifiedApi


def test_get_profiles(conan_api: ConanBaseUnifiedApi):
    profiles = conan_api.get_profiles()
    assert len(profiles) >= 1
    assert "default" in profiles


def test_get_profiles_path(conan_api: ConanBaseUnifiedApi):
    profiles_path = conan_api.get_profiles_path()
    assert (profiles_path / "default").is_file()


def test_get_profile_settings(conan_api: ConanBaseUnifiedApi):
    settings = conan_api.get_profile_settings("default")
    assert settings.get("os") == platform.system()
    assert settings.get("arch") == "x86_64"


def test_get_default_settings(conan_api: ConanBaseUnifiedApi):
    settings = conan_api.get_default_settings()
    assert settings.get("os") == platform.system()
    assert settings.get("arch") == "x86_64"


def test_conan_profile_name_alias_builder(conan_api: ConanBaseUnifiedApi):
    """ Test, that the build_conan_profile_name_alias returns human readable strings. """
    # check empty - should return a default name
    profile_name = conan_api.build_conan_profile_name_alias({})
    assert profile_name == "No Settings"

    # check a partial
    settings = {'os': 'Windows', 'arch': 'x86_64'}
    profile_name = conan_api.build_conan_profile_name_alias(settings)
    assert profile_name == "Windows_x64"

    # check windows
    WINDOWS_x64_VS16_SETTINGS = {'os': 'Windows', 'os_build': 'Windows', 'arch': 'x86_64',
                                 'arch_build': 'x86_64', 'compiler': 'Visual Studio',
                                 'compiler.version': '16', 'compiler.toolset': 'v142',
                                 'build_type': 'Release'}
    profile_name = conan_api.build_conan_profile_name_alias(WINDOWS_x64_VS16_SETTINGS)
    assert profile_name == "Windows_x64_vs16_v142_release"

    # check linux
    LINUX_X64_GCC7_SETTINGS = {'os': 'Linux', 'arch': 'x86_64', 'compiler': 'gcc',
                               'compiler.version': '7.4', 'build_type': 'Debug'}
    profile_name = conan_api.build_conan_profile_name_alias(LINUX_X64_GCC7_SETTINGS)
    assert profile_name == "Linux_x64_gcc7.4_debug"




from typing import List
import pytest
from test import (TEST_REF, TEST_REF_NO_SETTINGS, TEST_REF_OFFICIAL, TEST_REMOTE_NAME, 
                  test_ref_obj, test_ref_official_obj)
from test.conan_helper import conan_remove_ref

from conan_unified_api import conan_version
from conan_unified_api.types import ConanPkgRef, ConanRef
from conan_unified_api.unified_api import ConanBaseUnifiedApi


def test_info_simple(conan_api: ConanBaseUnifiedApi):
    # ref needs to be in a remote
    ref = ConanRef.loads(TEST_REF_OFFICIAL.split("@")[0])
    assert conan_api.get_conanfile_path(ref).exists()
    info = conan_api.info(ref)
    assert len(info) == 1
    if conan_version.major == 1:
        assert info[0].get("reference") == TEST_REF_OFFICIAL.split("@")[0]
        # assert info[0].get("binary_remote") == "local"
    elif conan_version.major == 2:  # binary_remote is usually None and reference does not work
        assert info[0].get("name") == ref.name


def test_info_transitive_reqs(conan_api: ConanBaseUnifiedApi):
    info = conan_api.info(TEST_REF_NO_SETTINGS)
    assert len(info) == 2

    if conan_version.major == 1:
        assert info[0].get("binary_remote") == TEST_REMOTE_NAME
        assert info[0].get("reference") == TEST_REF_NO_SETTINGS

        assert info[1].get("reference") == TEST_REF
    elif conan_version.major == 2:  # binary_remote is usually None and reference does not work
       assert info[0].get("name") == ConanRef.loads(TEST_REF_NO_SETTINGS).name
       assert info[1].get("name") == ConanRef.loads(TEST_REF).name

def test_conan_find_remote_pkg(conan_api: ConanBaseUnifiedApi):
    """
    Test, if search_package_in_remotes finds a package for the current system and the specified options.
    The function must find exactly one pacakge, which uses the spec. options and corresponds to the
    default settings.
    """
    conan_remove_ref(TEST_REF)
    default_settings = conan_api.get_default_settings()

    # check that inputing True as bool changes it in return to string 
    # - otherwise we cannot handle later comparisons.
    pkgs, remote = conan_api.find_best_matching_package_in_remotes(ConanRef.loads(TEST_REF),
                                                                   {"shared": True})
    assert remote == TEST_REMOTE_NAME
    assert len(pkgs) > 0
    pkg = pkgs[0]
    assert {"shared": True}.items() <= pkg.get("options", {}).items()

    for setting in default_settings:
        if setting in pkg.get("settings", {}).keys():
            if "compiler." in setting:  # don't evaluate comp. details
                continue
            assert default_settings[setting] in pkg.get("settings", {})[setting]


def test_conan_not_find_remote_pkg_wrong_opts(conan_api: ConanBaseUnifiedApi):
    """
    Test, if a wrong Option return causes an error.
    Empty list must be returned and the error be logged.
    """
    conan_remove_ref(TEST_REF)
    pkg, remote = conan_api.find_best_matching_package_in_remotes(ConanRef.loads(TEST_REF),
                                                                  {"BogusOption": "True"})
    assert not pkg


def test_search_for_all_packages(conan_api: ConanBaseUnifiedApi):
    """ Test, that an existing ref will be found in the remotes. """
    res = conan_api.search_recipe_all_versions_in_remotes(TEST_REF)
    assert len(res) >= 2
    assert TEST_REF in str(res)


@pytest.mark.parametrize("query, remote, expected",
                         [(TEST_REF, "local", [test_ref_obj]),
                          (TEST_REF, "all", [test_ref_obj]),
                          (TEST_REF_OFFICIAL, "local", []),  # does not work
                          ("example*", "local", [test_ref_official_obj, test_ref_obj]),
                         ],)
def test_search_recipes_in_remotes(conan_api: ConanBaseUnifiedApi, query: str, remote: str,
                                   expected: List["ConanRef"]):
    """ Test queries for conan search"""
    assert expected == conan_api.search_recipes_in_remotes(query, remote)


@pytest.mark.parametrize("ref", [TEST_REF, TEST_REF_OFFICIAL],)
def test_get_remote_pkg_from_id(conan_api: ConanBaseUnifiedApi, ref: str):
    """ Test finding the ConanPkg from the ConanPkgRef """
    pkg, _ = conan_api.find_best_matching_package_in_remotes(ref)
    assert len(pkg) >= 1
    assert pkg[0] == conan_api.get_remote_pkg_from_id(
        ConanPkgRef.loads(ref + ":" + pkg[0].get("id", "")))



import platform

from test import TEST_REF, TEST_REF_NO_SETTINGS, TEST_REMOTE_NAME
from test.conan_helper import conan_remove_ref

from conan_unified_api import conan_version
from conan_unified_api.types import ConanRef
from conan_unified_api.unified_api import ConanBaseUnifiedApi


def test_get_path_or_install(conan_api: ConanBaseUnifiedApi):
    """
    Test, if get_package installs the package and returns the path and check it again.
    The bin dir in the package must exist (indicating it was correctly downloaded)
    """
    dir_to_check = "bin"
    conan_remove_ref(TEST_REF)

    # Gets package path / installs the package
    id, package_folder = conan_api.get_path_with_auto_install(ConanRef.loads(TEST_REF))
    assert (package_folder / dir_to_check).is_dir()
    # check again for already installed package
    id, package_folder = conan_api.get_path_with_auto_install(ConanRef.loads(TEST_REF))
    assert (package_folder / dir_to_check).is_dir()


def test_get_path_or_install_manual_options(conan_api: ConanBaseUnifiedApi):
    """
    Test, if a package with options can install.
    The actual installaton must not return an error and non given options be merged with default options.
    """
    # This package has an option "shared" and is fairly small.
    conan_remove_ref(TEST_REF)
    id, package_folder = conan_api.get_path_with_auto_install(
        ConanRef.loads(TEST_REF), {"shared": "True"})
    if platform.system() == "Windows":
        assert (package_folder / "bin" / "python.exe").is_file()
    elif platform.system() == "Linux":
        assert (package_folder / "bin" / "python").is_file()


def test_install_with_any_settings(mocker, capfd, conan_api: ConanBaseUnifiedApi):
    """
    Test, if a package with <setting>=Any flags can install
    The actual installaton must not return an error.
    """
    if conan_version.major == 2:  # TODO create package for it
        return
    # mock the remote response
    conan_remove_ref(TEST_REF)
    # Create the "any" package
    assert conan_api.install_package(ConanRef.loads(TEST_REF), {
        'id': '325c44fdb228c32b3de52146f3e3ff8d94dddb60', 'options': {},
        'settings': {'arch_build': 'any', 'os_build': 'Linux', "build_type": "ANY"},
        'requires': [], 'outdated': False}, False, TEST_REMOTE_NAME)
    captured = capfd.readouterr()
    assert "ERROR" not in captured.err
    assert "Cannot install package" not in captured.err

def test_install_compiler_no_settings(conan_api: ConanBaseUnifiedApi, capfd):
    """
    Test, if a package with no settings at all can install
    The actual installaton must not return an error.
    """
    if conan_version.major == 2:  # FIXME: test does not work yet!
        return
    ref = TEST_REF_NO_SETTINGS
    conan_remove_ref(ref)
    capfd.readouterr() # remove can result in error message - clear

    id, package_folder = conan_api.get_path_with_auto_install(ConanRef.loads(ref))
    assert (package_folder / "bin").is_dir()
    captured = capfd.readouterr()
    assert "ERROR" not in captured.err
    assert "Can't find a matching package" not in captured.err
    conan_remove_ref(ref)



def test_conan_get_conan_buildinfo(conan_api: ConanBaseUnifiedApi):
    """
    Check, that get_conan_buildinfo actually retrieves as a string for the linux pkg 
    This exectues an install under the hood, thus the category
    """
    if conan_version.major == 2: # not implemented yet
        return
    LINUX_X64_GCC9_SETTINGS = {'os': 'Linux', 'arch': 'x86_64', 'compiler': 'gcc',
                               "compiler.libcxx": "libstdc++11", 'compiler.version': '9', 'build_type': 'Release'}
    buildinfo = conan_api.get_conan_buildinfo(
        ConanRef.loads(TEST_REF), LINUX_X64_GCC9_SETTINGS)
    assert "USER_example" in buildinfo
    assert "ENV_example" in buildinfo




from importlib.metadata import distribution
import os
from pathlib import Path
from packaging.version import Version

INVALID_PATH_VALUE = "/INVALIDPATH_6a998a6f-c544-4a32-8414-4dd317d905a3"
invalid_path = Path(INVALID_PATH_VALUE)
DEBUG_LEVEL = int(os.getenv("CONAN_UNIFIED_API_DEBUG_LEVEL", "0"))
CONAN_LOG_PREFIX = "CONAN - "

conan_pkg_info = distribution("conan")
conan_version = Version(conan_pkg_info.version)




import pytest
from test import TEST_REF, TEST_REF_OFFICIAL, TEST_REMOTE_NAME, test_ref_obj, test_ref_official_obj
from test.conan_helper import conan_install_ref, conan_remove_ref, get_profiles

from conan_unified_api.types import ConanPkgRef, ConanRef
from conan_unified_api import conan_version
from conan_unified_api import ConanUnifiedApi
from conan_unified_api.common import ConanUnifiedApi

def test_inspect(conan_api: ConanUnifiedApi):
    inspect = conan_api.inspect(TEST_REF)
    assert inspect.get("name") == ConanRef.loads(TEST_REF).name
    assert inspect.get("generators") == ("CMakeDeps", "CMakeToolchain")

    # objects are not identical -> compare arbitrary values 
    inspect == conan_api.inspect(test_ref_obj)
    assert inspect.get("name") == ConanRef.loads(TEST_REF).name
    assert inspect.get("generators") == ("CMakeDeps", "CMakeToolchain")

    inspect = conan_api.inspect(TEST_REF, ["no_copy_source"], TEST_REMOTE_NAME)
    assert inspect["no_copy_source"] == True


def test_alias(conan_api: ConanUnifiedApi):
    """ Test that alias creates the ref locally -> will only have export folder """
    if conan_version.major == 2: # skip for conan 2
        return
    alias_ref = "example/(1.1.1)@user/new_channel"
    try:
        conan_api.alias(alias_ref, TEST_REF)
        assert conan_api.get_export_folder(ConanRef.loads(alias_ref)).exists()
    finally:
        conan_remove_ref(alias_ref)

def test_conan_find_local_pkg(conan_api: ConanUnifiedApi):
    """
    Test, if get_package installs the package and returns the path and check it again.
    The bin dir in the package must exist (indicating it was correctly downloaded)
    """
    conan_remove_ref(TEST_REF)
    conan_install_ref(TEST_REF)
    pkgs = conan_api.find_best_matching_packages(ConanRef.loads(TEST_REF))
    assert len(pkgs) == 1 # default options are filtered


def test_get_export_folder(conan_api: ConanUnifiedApi):
    conan_install_ref(TEST_REF)
    assert (conan_api.get_export_folder(TEST_REF) / "conanfile.py").exists()
    assert (conan_api.get_export_folder(test_ref_obj) / "conanfile.py").exists()


def test_get_conanfile_path(conan_api: ConanUnifiedApi):
    conanfile_path = conan_api.get_conanfile_path(TEST_REF)
    assert conanfile_path.is_file()
    assert conanfile_path.name == "conanfile.py"
    assert conanfile_path == conan_api.get_conanfile_path(test_ref_obj)


def test_get_all_local_refs(conan_api: ConanUnifiedApi):
    conan_install_ref(TEST_REF)
    conan_install_ref(TEST_REF_OFFICIAL)

    refs = conan_api.get_all_local_refs()
    assert test_ref_obj in refs
    assert test_ref_official_obj in refs


def test_get_local_pkg_from_id(conan_api: ConanUnifiedApi):
    conan_install_ref(TEST_REF)
    pkgs = conan_api.get_local_pkgs_from_ref(TEST_REF)

    pkg = conan_api.get_local_pkg_from_id(ConanPkgRef(TEST_REF, pkgs[0].get("id")))
    
    assert pkg == pkgs[0]

def test_get_local_pkg_from_path(conan_api: ConanUnifiedApi):
    pass


def test_get_options_with_default_values(conan_api: ConanUnifiedApi):
    available_options, default_options = conan_api.get_options_with_default_values(test_ref_obj)
    assert conan_api._are_option_compatible(available_options, {'shared': [
        'True', 'False'], 'fPIC2': ['True', 'False'], 'variant': ['ANY']})
    conan_api._are_option_compatible(
        default_options, {"shared": True, 'fPIC2': True, "variant": "var1"})

def test_get_local_pkgs_from_ref(conan_api: ConanUnifiedApi):
    # install all packages
    for profile in get_profiles():
        for option in ["True", "False"]:
            conan_install_ref(TEST_REF, "-o shared=" + option, profile)
    pkgs = conan_api.get_local_pkgs_from_ref(TEST_REF)
    if conan_version.major == 1:
        assert len(pkgs) == 4
    else:
        assert len(pkgs) >= 2  # TODO this seems to be a bug in conanV2


def test_get_package_folder(conan_api: ConanUnifiedApi):
    pkgs = conan_api.get_local_pkgs_from_ref(TEST_REF)
    pkg_path = conan_api.get_package_folder(TEST_REF, pkgs[0].get("id", ""))
    assert pkg_path.exists() # TODO better check


def test_remove_reference(conan_api: ConanUnifiedApi):
    id, path = conan_api.get_path_with_auto_install(TEST_REF)
    assert path.exists()

    conan_api.remove_reference(TEST_REF)
    assert not path.exists()


@pytest.mark.parametrize("ref, option_key, option_value",
                         [(TEST_REF, "shared", True),
                          (TEST_REF_OFFICIAL, "", None),
                          ],)
def test_find_best_matching_local_package(conan_api: ConanUnifiedApi, ref: str, 
                                          option_key: str, option_value):
    """ Test find a package in the local cache """
    option = None
    if option_key:
        option = {option_key: option_value}
    conan_api.get_path_with_auto_install(test_ref_obj, option)
    matching_pkg = conan_api.find_best_matching_local_package(TEST_REF, option)
    assert matching_pkg.get("id")
    assert matching_pkg.get("options")
    assert matching_pkg.get("settings")

    if option_key:
        assert matching_pkg.get("options", {})[option_key] == option_value
    

def test_get_best_matching_local_package_path(conan_api: ConanUnifiedApi):
    # TODO add conan options as test parameter
    conan_install_ref(TEST_REF)
    
    id, path = conan_api.get_best_matching_local_package_path(test_ref_obj)
    
    pkg = conan_api.get_local_pkg_from_path(test_ref_obj, path)

    assert path.exists()
    assert pkg.get("id") == id



import logging
import os
import platform
from contextlib import redirect_stderr, redirect_stdout
from pathlib import Path
from tempfile import gettempdir
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Sequence, Tuple, Union
from unittest.mock import patch

from typing_extensions import Self

from conan_unified_api.base.helper import create_key_value_pair_list, delete_path, save_sys_path
from conan_unified_api.base.typing import SignatureCheckMeta

try:
    from contextlib import chdir
except ImportError:
    from contextlib_chdir import chdir  # type: ignore

from .common import ConanUnifiedApi
from .types import (
    ConanAvailableOptions,
    ConanException,
    ConanOptions,
    ConanPackageId,
    ConanPackagePath,
    ConanPkg,
    ConanPkgRef,
    ConanRef,
    ConanSettings,
    EditablePkg,
    LoggerWriter,
    Remote,
)

if TYPE_CHECKING:
    from conans.client.conan_api import ClientCache, ConanAPIV1

    from .cache.conan_cache import ConanInfoCache

from conan_unified_api.base import CONAN_LOG_PREFIX, INVALID_PATH_VALUE, invalid_path

current_path = Path(__file__).parent


class ConanApi(ConanUnifiedApi, metaclass=SignatureCheckMeta):
    """Wrapper around ConanAPIV1"""

    def __init__(
        self,
        init: bool = True,
        logger: Optional[logging.Logger] = None,
        mute_logging: bool = False,
    ):
        self._conan: ConanAPIV1
        self._client_cache: ClientCache
        self._short_path_root = Path("Unknown")
        self.info_cache: ConanInfoCache
        super().__init__(init, logger, mute_logging)

    def init_api(self) -> Self:
        """Instantiate the internal Conan api."""
        from conans.client.conan_api import ConanAPIV1, ConanApp, UserIO
        from conans.client.output import ConanOutput

        self._fix_editable_file()

        self._conan = ConanAPIV1(
            output=ConanOutput(
                LoggerWriter(self.logger.info, CONAN_LOG_PREFIX),
                LoggerWriter(self.logger.error, CONAN_LOG_PREFIX),
            )
        )
        self._conan.user_io = UserIO(
            out=ConanOutput(
                LoggerWriter(self.logger.info, CONAN_LOG_PREFIX),
                LoggerWriter(self.logger.error, CONAN_LOG_PREFIX),
            )
        )
        self._conan.create_app()
        self._conan.user_io.disable_input()  # error on inputs - nowhere to enter
        if self._conan.app:
            self._client_cache = self._conan.app.cache
        else:
            raise NotImplementedError

        def create_app(self, quiet_output=None, force: bool = False):
            if self.app and not force:
                return self.app
            self.app = ConanApp(  # noqa: RET503
                self.cache_folder,
                self.user_io,
                self.http_requester,
                self.runner,
                quiet_output=quiet_output,
            )

        patch("conans.client.conan_api.ConanAPIV1.create_app", create_app).start()

        # don't hang on startup
        try:  # use try-except because of Conan 1.24 envvar errors in tests
            self.remove_locks()
        except Exception as e:
            self.logger.debug(str(e), exc_info=True)
        from .cache.conan_cache import ConanInfoCache

        self.info_cache = ConanInfoCache(current_path, self.get_all_local_refs())
        self.logger.debug("Initialized Conan V1 API wrapper")

        return self

    def _fix_editable_file(self) -> None:
        """Ensure editables json is valid (Workaround for empty json bug)
        Must work without using ConanAPIV1, because it can't be called without this
        """
        from conans.client.cache.editable import EDITABLE_PACKAGES_FILE
        from conans.paths import get_conan_user_home  # use internal fnc

        try:
            editable_file_path = Path(get_conan_user_home()) / ".conan" / EDITABLE_PACKAGES_FILE
            if not editable_file_path.exists():
                editable_file_path.write_text("{}")
            content = editable_file_path.read_text()
            if not content:
                editable_file_path.write_text("{}")
        except Exception:
            self.logger.debug("Reinit editable file package")

    ### General commands ###

    def remove_locks(self) -> None:
        self._conan.remove_locks()
        self.logger.debug("Removed Conan cache locks.")

    def info(self, conan_ref: Union[ConanRef, str]) -> List[Dict[str, Any]]:
        with save_sys_path():  # can change path or run arbitrary code and thus break things
            canonical_ref = self.generate_canonical_ref(conan_ref)
            deps_graph = self._conan.info(canonical_ref)
            # ugly hack, but sadly no other way to do this
            from conans.client.command import CommandOutputer

            infos: List[Dict[str, Any]] = CommandOutputer(
                self._conan.out, self._client_cache
            )._grab_info_data(deps_graph[0], grab_paths=True)
            # insert original ref as 0th element
            own_info = {}
            for info in infos:
                if canonical_ref == self.generate_canonical_ref(
                    ConanRef.loads(info.get("reference", ""))
                ):
                    own_info = info
                    break
            if not own_info:
                msg = "Can't find own reference in info list!"
                raise ConanException(msg)
            infos.remove(own_info)
            infos.insert(0, own_info)
            return infos

    def inspect(
        self,
        conan_ref: Union[ConanRef, str],
        attributes: Sequence[str] = (),
        remote_name: Optional[str] = None,
    ) -> Dict[str, Any]:
        with save_sys_path():  # can change path or run arbitrary code and thus break things
            # cast from ordered dict
            return dict(
                self._conan.inspect(
                    str(conan_ref), attributes=attributes, remote_name=remote_name
                )
            )

    def alias(
        self, conan_ref: Union[ConanRef, str], conan_target_ref: Union[ConanRef, str]
    ) -> None:
        self._conan.export_alias(str(conan_ref), conan_target_ref)

    def get_profiles(self) -> List[str]:
        return self._conan.profile_list()

    def get_profile_settings(self, profile_name: str) -> ConanSettings:
        profile = self._conan.read_profile(profile_name)
        if not profile:
            return {}
        return dict(profile.settings)

    def get_default_settings(self) -> ConanSettings:
        default_profile = self._client_cache.default_profile
        if not default_profile:
            return {}
        return dict(default_profile.settings)

    # user_name, authenticated
    def get_remote_user_info(self, remote_name: str) -> Tuple[str, bool]:
        user_info = self._conan.users_list(remote_name).get("remotes", {})
        if len(user_info) < 1:
            return ("", False)
        try:
            return (
                str(user_info[0].get("user_name", "")),
                user_info[0].get("authenticated", False),
            )
        except Exception:
            self.logger.warning("Can't get user info for %s", remote_name)
            return ("", False)

    def get_config_file_path(self) -> Path:
        return Path(self._client_cache.conan_conf_path)

    def get_config_entry(self, config_name: str) -> Optional[Any]:
        """Will always return str, but 2 will return correct type TODO: implement this?"""
        try:
            self._conan.out._stream.disabled = True  # type: ignore
            config_value = self._conan.config_get(config_name)  # mute this output
            self._conan.out._stream.disabled = False  # type: ignore
            return config_value
        except Exception:
            try:
                config_entry_suffix = config_name.split(".")[1]
                config_value = self._client_cache.config.env_vars.get(
                    "CONAN_" + config_entry_suffix.upper(), None
                )
                return config_value
            except Exception:
                return None

    def get_revisions_enabled(self) -> bool:
        return self._client_cache.config.revisions_enabled

    def get_settings_file_path(self) -> Path:
        return Path(self._client_cache.settings_path)

    def get_profiles_path(self) -> Path:
        return Path(str(self._client_cache.default_profile_path)).parent

    def get_user_home_path(self) -> Path:
        return Path(self._client_cache.cache_folder)

    def get_storage_path(self) -> Path:
        return Path(str(self._client_cache.store))

    def get_editable_references(self) -> List[ConanRef]:
        try:
            self._conan.create_app(force=True)  # type: ignore # need to possibly reload editables
            return list(map(ConanRef.loads, self._conan.editable_list().keys()))
        except Exception:
            self._fix_editable_file()  # to not crash conan without this
            return []

    def get_editable(self, conan_ref: Union[ConanRef, str]) -> Optional[EditablePkg]:
        conan_ref = self.conan_ref_from_reflike(conan_ref)
        self._conan.create_app(force=True)  # type: ignore # need to possibly reload editables

        editable_dict = self._conan.editable_list().get(str(conan_ref), {})
        if not editable_dict:
            return None
        return EditablePkg(
            str(conan_ref),
            editable_dict.get("path", INVALID_PATH_VALUE),
            editable_dict.get("output_folder"),
        )

    def get_editables_package_path(self, conan_ref: Union[ConanRef, str]) -> Path:
        self._conan.create_app(force=True)  # type: ignore # need to possibly reload editables
        editable_dict = self._conan.editable_list().get(str(conan_ref), {})
        return Path(str(editable_dict.get("path", INVALID_PATH_VALUE)))

    def get_editables_output_folder(self, conan_ref: Union[ConanRef, str]) -> Optional[Path]:
        self._conan.create_app(force=True)  # type: ignore # need to possibly reload editables
        editable_dict = self._conan.editable_list().get(str(conan_ref), {})
        output_folder = editable_dict.get("output_folder")
        if not output_folder:
            return None
        return Path(str(output_folder))

    def add_editable(
        self,
        conan_ref: Union[ConanRef, str],
        path: Union[Path, str],
        output_folder: Union[Path, str],
    ) -> bool:
        try:
            self._conan.editable_add(str(path), str(conan_ref), None, str(output_folder), None)
        except Exception as e:
            self.logger.error("Error adding editable: %s", str(e))
            return False
        return True

    def remove_editable(self, conan_ref: Union[ConanRef, str]) -> bool:
        try:
            self._conan.editable_remove(str(conan_ref))
        except Exception as e:
            self.logger.error("Error removing editable: %s", str(e))
            return False
        return True

    def get_short_path_root(self) -> Path:
        # only need to get once
        if self._short_path_root.exists() or platform.system() != "Windows":
            return self._short_path_root
        short_home = self._client_cache.config.short_paths_home
        if not short_home:
            drive = os.path.splitdrive(self._client_cache.cache_folder)[0].upper()
            short_home = os.path.join(drive, os.sep, ".conan")
        else:
            short_home = str(short_home)
        short_home_path = Path(short_home)
        Path.mkdir(short_home_path, exist_ok=True)
        return short_home_path

    def get_package_folder(self, conan_ref: Union[ConanRef, str], package_id: str) -> Path:
        if not package_id:  # will give the base path otherwise
            return invalid_path
        conan_ref = self.conan_ref_from_reflike(conan_ref)
        try:
            layout = self._client_cache.package_layout(conan_ref)
            return Path(layout.package(ConanPkgRef(conan_ref, package_id)))
        except Exception:  # gotta catch 'em all!
            return invalid_path

    def get_export_folder(self, conan_ref: Union[ConanRef, str]) -> Path:
        layout = self._client_cache.package_layout(self.conan_ref_from_reflike(conan_ref))
        if layout:
            return Path(layout.export())
        return invalid_path

    def get_conanfile_path(self, conan_ref: Union[ConanRef, str]) -> Path:
        try:
            if conan_ref not in self.get_all_local_refs():
                self.inspect(conan_ref)
            layout = self._client_cache.package_layout(self.conan_ref_from_reflike(conan_ref))
            if layout:
                return Path(layout.conanfile())
        except Exception as e:
            msg = f"Can't get conanfile: {e!s}"
            raise ConanException(msg) from e
        return invalid_path

    # Remotes

    def get_remotes(self, include_disabled: bool = False) -> List[Remote]:
        remotes = []
        try:
            if include_disabled:
                remotes = self._conan.remote_list()
            else:
                remotes = self._client_cache.registry.load_remotes().values()
        except Exception as e:
            msg = f"Error while reading remotes file: {e!s}"
            raise ConanException(msg) from e
        return remotes  # type: ignore

    def add_remote(self, remote_name: str, url: str, verify_ssl: bool) -> None:
        self._conan.remote_add(remote_name, url, verify_ssl)

    def rename_remote(self, remote_name: str, new_name: str) -> None:
        self._conan.remote_rename(remote_name, new_name)

    def remove_remote(self, remote_name: str) -> None:
        self._conan.remote_remove(remote_name)

    def disable_remote(self, remote_name: str, disabled: bool) -> None:
        self._conan.remote_set_disabled_state(remote_name, disabled)

    def update_remote(
        self, remote_name: str, url: str, verify_ssl: bool, index: Optional[int] = None
    ) -> None:
        self._conan.remote_update(remote_name, url, verify_ssl, index)

    def login_remote(self, remote_name: str, user_name: str, password: str) -> None:
        self._conan.authenticate(user_name, password, remote_name)

    ### Install related methods ###

    def install_reference(
        self,
        conan_ref: Union[ConanRef, str],
        conan_settings: Optional[ConanSettings] = None,
        conan_options: Optional[ConanOptions] = None,
        profile: str = "",
        update: bool = True,
        generators: Sequence[str] = (),
        remote_name: Optional[str] = None,
    ) -> Tuple[ConanPackageId, ConanPackagePath]:
        package_id = ""
        if conan_options is None:
            conan_options = {}
        if conan_settings is None:
            conan_settings = {}
        options_list = create_key_value_pair_list(conan_options)
        settings_list = create_key_value_pair_list(conan_settings)
        install_message = (
            f"Installing '{conan_ref!s}' with profile: {profile}, "
            f"settings: {settings_list!s}, "
            f"options: {options_list!s} and update={update}\n"
        )
        self.logger.info(install_message)
        profile_names = None
        if profile:
            profile_names = [profile]
        try:
            # Try to redirect custom streams in conanfile, to avoid missing flush method
            devnull = open(os.devnull, "w")
            # also spoof os.terminal_size(
            spoof_size = os.terminal_size([80, 20])
            patched_tersize = patch("os.get_terminal_size")
            with redirect_stdout(devnull), redirect_stderr(devnull):
                mock = patched_tersize.start()
                mock.return_value = spoof_size

                infos = self._conan.install_reference(
                    conan_ref,
                    settings=settings_list,
                    options=options_list,
                    update=update,
                    profile_names=profile_names,
                    generators=generators,
                    remote_name=remote_name,
                )

                patched_tersize.stop()
            if not infos.get("error", True):
                package_id = (
                    infos.get("installed", [{}])[0].get("packages", [{}])[0].get("id", "")
                )
            self.logger.info("Installation of '%s' finished", str(conan_ref))
            return package_id, self.get_package_folder(conan_ref, package_id)
        except ConanException as e:
            msg = f"Can't install reference {conan_ref}': {e!s}"
            raise ConanException(msg) from e

    def get_conan_buildinfo(
        self,
        conan_ref: Union[ConanRef, str],
        conan_settings: ConanSettings,
        conan_options: Optional[ConanOptions] = None,
    ) -> str:
        # install ref to temp dir and use generator
        temp_path = Path(gettempdir()) / "cal_cuild_info"
        temp_path.mkdir(parents=True, exist_ok=True)
        generated_file = temp_path / "conanbuildinfo.txt"
        # clean up possible last run
        delete_path(generated_file)

        # use cli here, API cannnot do job easily and we wan to parse the file output
        with chdir(temp_path):
            self.install_reference(
                conan_ref,
                conan_settings=conan_settings,
                conan_options=conan_options,
                generators=["txt"],
            )
        content = ""
        try:
            content = generated_file.read_text()
        except Exception as e:
            msg = f"Can't read conanbuildinfo.txt for '{conan_ref}': {e!s}"
            raise ConanException(msg) from e
        return content

    def get_options_with_default_values(
        self, conan_ref: Union[ConanRef, str], remote_name: Optional[str] = None
    ) -> Tuple[ConanAvailableOptions, ConanOptions]:
        # this calls external code of the recipe
        default_options = {}
        available_options = {}
        try:
            inspect = self.inspect(
                self.generate_canonical_ref(conan_ref), remote_name=remote_name
            )
            default_options = inspect.get("default_options", {})
            available_options = inspect.get("options", {})
        except Exception as e:
            self.logger.error(
                "Error while getting default options for %s: %s", str(conan_ref), str(e)
            )
        return available_options, default_options

    # Local References and Packages

    def remove_reference(self, conan_ref: Union[ConanRef, str], pkg_id: str = "") -> None:
        pkg_ids = [pkg_id] if pkg_id else None
        self._conan.remove(str(conan_ref), packages=pkg_ids, force=True)

    def get_all_local_refs(self) -> List[ConanRef]:
        return self._client_cache.all_refs()  # type: ignore

    def get_local_pkgs_from_ref(self, conan_ref: Union[ConanRef, str]) -> List[ConanPkg]:
        result: List[ConanPkg] = []
        response = {}
        try:
            response = self._conan.search_packages(self.generate_canonical_ref(conan_ref))
        except Exception as e:
            msg = f"Error while getting local packages for recipe: {e!s}"
            raise ConanException(msg) from e
        if not response.get("error", True):
            try:
                result = (
                    response.get("results", [{}])[0].get("items", [{}])[0].get("packages", [{}])
                )
            except Exception as e:
                msg = f"Received invalid package response format for {conan_ref!s}"
                raise ConanException(msg) from e
        return result

    # Remote References and Packages

    def search_recipes_in_remotes(self, query: str, remote_name: str = "all") -> List[ConanRef]:
        result_recipes: List[ConanRef] = []
        remote_results = []
        try:
            # no query possible with pattern
            remote_results = self._conan.search_recipes(
                query, remote_name=remote_name, case_sensitive=False
            ).get("results", None)
        except Exception as e:
            msg = f"Error while searching for recipe: {e!s}"
            raise ConanException(msg) from e
        if not remote_results:
            return result_recipes

        for remote_search_res in remote_results:
            result_recipes += [
                ConanRef.loads(item.get("recipe", {}).get("id", ""))
                for item in remote_search_res.get("items", [])
            ]
        result_recipes = list(set(result_recipes))  # make unique
        result_recipes.sort()
        self.info_cache.update_remote_package_list(result_recipes)
        return result_recipes

    def search_recipe_all_versions_in_remotes(
        self, conan_ref: Union[ConanRef, str]
    ) -> List[ConanRef]:
        conan_ref = self.conan_ref_from_reflike(conan_ref)

        remote_results: List[Dict[str, Any]] = []
        local_results: List[Dict[str, Any]] = []
        try:
            # no query possible with pattern
            remote_results = self._conan.search_recipes(
                f"{conan_ref.name}/*@*/*", remote_name="all"
            ).get("results", None)
        except Exception as e:
            self.logger.warning(str(e))
        try:
            local_results = self._conan.search_recipes(
                f"{conan_ref.name}/*@*/*", remote_name=None
            ).get("results", None)
        except Exception as e:
            self.logger.warning(str(e))
            return []

        res_list: List[ConanRef] = []
        for remote_search_res in local_results + remote_results:
            res_list += [
                ConanRef.loads(item.get("recipe", {}).get("id", ""))
                for item in remote_search_res.get("items", [])
            ]
        res_list = list(set(res_list))  # make unique
        res_list.sort()
        # update cache
        self.info_cache.update_remote_package_list(res_list)
        return res_list

    def get_remote_pkgs_from_ref(
        self,
        conan_ref: Union[ConanRef, str],
        remote_name: Optional[str],
        query: Optional[str] = None,
    ) -> List[ConanPkg]:
        conan_ref = self.conan_ref_from_reflike(conan_ref)
        found_pkgs: List[ConanPkg] = []
        try:
            search_results = self._conan.search_packages(
                conan_ref.full_str(), query=query, remote_name=remote_name
            ).get("results", None)
        except Exception as e:
            self.logger.debug(
                "Error while searching for %s in %s: %s", conan_ref, remote_name, str(e)
            )
            return []
        if search_results:
            found_pkgs = search_results[0].get("items")[0].get("packages")

        return found_pkgs



# Typing helpers for enforced interfaces

import inspect
from types import FunctionType

from . import DEBUG_LEVEL


class SignatureMismatchError(Exception):
    pass


class SignatureCheckMeta(type):
    def __new__(cls, name, base_classes, methods):  # noqa: ANN001
        """
        For each method, check if any base class already defined a
        method with that name. If so, make sure the signatures are the same.
        """
        if DEBUG_LEVEL < 1:
            return type(name, base_classes, methods)
        for method_name in methods:
            # allow for different constructors through factories
            if method_name == "__init__":
                continue
            method = methods[method_name]

            if not isinstance(method, FunctionType):
                continue
            for base_class in base_classes:
                try:
                    base_method = getattr(base_class, method_name)
                    base_argspec = inspect.getfullargspec(base_method)
                    method_argspec = inspect.getfullargspec(method)
                    if method_argspec != base_argspec:
                        raise SignatureMismatchError(
                            method_name + "\n"
                            "Expected: " + str(base_argspec) + "\n"
                            "Actual: " + str(method_argspec),
                        )
                except AttributeError:  # noqa: PERF203
                    # method was not defined in base class, skip
                    continue
        return type(name, base_classes, methods)



import logging
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from typing import Self


class Logger(logging.Logger):
    """
    Singleton instance for the global logger
    """

    _instance: Optional[logging.Logger] = None
    formatter = logging.Formatter(r"%(levelname)s: %(message)s")

    def __new__(cls) -> "Self":
        if cls._instance is None:
            # the user excepts a logger
            cls._instance = cls._init_logger()
        return cls._instance  # type: ignore

    def __init__(self) -> None:
        return None

    @classmethod
    def _init_logger(cls) -> logging.Logger:
        """Set up format and a debug level and register console logger."""
        # restrict root logger
        root = logging.getLogger()
        root.setLevel(logging.ERROR)

        from conan_unified_api import PKG_NAME
        from conan_unified_api.base import DEBUG_LEVEL

        logger = logging.getLogger(PKG_NAME)
        logger.setLevel(logging.DEBUG)
        log_debug_level = logging.INFO
        if DEBUG_LEVEL > 0:
            log_debug_level = logging.DEBUG

        console_handler = logging.StreamHandler()
        console_handler.setLevel(log_debug_level)

        console_handler.setFormatter(cls.formatter)

        logger.addHandler(console_handler)

        # otherwise messages appear twice
        logger.propagate = False
        return logger



from __future__ import annotations

import os
import platform
import pprint
from contextlib import redirect_stderr, redirect_stdout
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Dict, List, Literal, Optional, TypedDict, Union

from typing_extensions import TypeAlias

from conan_unified_api import conan_version

if conan_version.major == 1:
    from conans.errors import ConanException  # noqa: F401
    from conans.model.ref import ConanFileReference, PackageReference
    from conans.paths.package_layouts.package_editable_layout import (
        PackageEditableLayout,  # noqa: F401
    )

    if platform.system() == "Windows":
        from conans.util.windows import CONAN_LINK, CONAN_REAL_PATH
else:
    try:
        from conans.model.package_ref import PkgReference
        from conans.model.recipe_ref import RecipeReference as ConanFileRef
    except ImportError:  # try again for versions where the import has a circular dependency
        from conans.model.package_ref import PkgReference
        from conans.model.recipe_ref import RecipeReference as ConanFileRef
    try:
        from conan.errors import ConanException
    except ImportError:  # until conan version 2.?
        from conans.errors import ConanException  # noqa: F401

    class PackageReference(PkgReference):
        """Compatibility class for changed package_id attribute"""

        ref: ConanRef

        @property
        def id(self) -> str:
            return self.package_id

        @staticmethod
        def loads(text: str) -> ConanPkgRef:
            pkg_ref = PkgReference.loads(text)
            return PackageReference(
                pkg_ref.ref, pkg_ref.package_id, pkg_ref.revision, pkg_ref.timestamp
            )

    class ConanFileReference(ConanFileRef):
        """Compatibility class for validation in loads method"""

        name: str
        version: str
        user: Optional[str]
        channel: Optional[str]

        @staticmethod
        def loads(text: str, validate: bool = True) -> ConanRef:  # type: ignore
            # add back support for @_/_ canonical refs to handle this uniformly
            # Simply remove it before passing it to ConanFileRef
            if text.endswith("@_/_"):
                text = text.replace("@_/_", "")
            ref: ConanRef = ConanFileRef().loads(text)  # type: ignore
            if validate:
                # validate_ref creates an own output stream which can't log to console
                # if it is running as a gui application
                with open(os.devnull, "w") as devnull:
                    with redirect_stdout(devnull), redirect_stderr(devnull):
                        ref.validate_ref(allow_uppercase=True)
            return ref


@dataclass
class Remote:
    name: str
    url: str
    verify_ssl: bool
    disabled: bool
    allowed_packages: Optional[List[str]] = None


ConanRef: TypeAlias = ConanFileReference
ConanPkgRef: TypeAlias = PackageReference
ConanOptions: TypeAlias = Dict[str, Any]
ConanAvailableOptions: TypeAlias = Dict[str, Union[List[Any], Literal["ANY"]]]
ConanSettings: TypeAlias = Dict[str, str]
ConanPackageId: TypeAlias = str
ConanPackagePath: TypeAlias = Path


class ConanPkg(TypedDict, total=False):
    """Dummy class to type conan returned package dicts"""

    id: ConanPackageId
    options: ConanOptions
    settings: ConanSettings
    requires: List[Any]
    outdated: bool


@dataclass
class EditablePkg:
    conan_ref: str
    path: str  # path to conanfile or folder
    output_folder: Optional[str]


def pretty_print_pkg_info(pkg_info: ConanPkg) -> str:
    return pprint.pformat(pkg_info).translate({ord("{"): None, ord("}"): None, ord("'"): None})


class LoggerWriter:
    """
    Dummy stream to log directly to a logger object, when writing in the stream.
    Used to redirect custom stream from Conan.
    Adds a prefix to do some custom formatting in the Logger.
    """

    disabled = False

    def __init__(self, level: Callable, prefix: str):
        self.level = level
        self._prefix = prefix

    def write(self, message: str) -> None:
        if self.disabled:
            return
        if message != "\n":
            self.level(self._prefix + message.strip("\n"))

    def flush(self) -> None:
        """For interface compatiblity"""




import os
import platform
import tempfile
from pathlib import Path

from conan_unified_api import conan_version
from conan_unified_api.base.helper import str2bool
from conan_unified_api.unified_api import ConanBaseUnifiedApi


def test_get_settings_file_path(conan_api: ConanBaseUnifiedApi):
    settings_path = conan_api.get_settings_file_path()
    assert settings_path.name == "settings.yml"
    assert settings_path.is_file()


def test_get_config_file_path(conan_api: ConanBaseUnifiedApi):
    config_path = conan_api.get_config_file_path()
    assert config_path.suffix == ".conf"
    assert config_path.is_file()


def test_get_config_entry(conan_api: ConanBaseUnifiedApi):
    if conan_version.major == 1:
        config_entry_name = "general.non_interactive" # set by fixture
    else:
        # no adding config api yet, so hack it 
        config = conan_api.get_config_file_path().read_text()
        # remove commented out lines
        configs = {}
        for line in config.split("\n"):
            if not line.lstrip().startswith("#") and line:
                name = line.split("=")[0].strip()
                value = line.split("=")[1].strip()
                configs[name] = value
        config_entry_name = "core:non_interactive"
        if config_entry_name not in configs:
            config += f"\n{config_entry_name}=True\n"
            conan_api.get_config_file_path().write_text(config)
    conan_api.init_api() # reads config only on init
    entry_value = conan_api.get_config_entry(config_entry_name)
    assert entry_value is not None

    if conan_version.major == 1:
        assert str2bool(entry_value)
    else:
        assert entry_value == True # want to know if really True and not some truthy value


def test_get_revisions_enabled(conan_api: ConanBaseUnifiedApi):
    entry_value = conan_api.get_revisions_enabled()
    assert entry_value is not None
    assert entry_value


def test_get_user_home_path(conan_api: ConanBaseUnifiedApi):
    path = conan_api.get_user_home_path()
    assert path.is_dir()
    assert (path / "settings.yml").exists() # ok for both Conan 1 and 2


def test_get_storage_path(conan_api: ConanBaseUnifiedApi):
    path = conan_api.get_storage_path()
    assert path.is_dir()
    # TODO: Extend test a little bit...


def test_conan_short_path_root(conan_api: ConanBaseUnifiedApi):
    """ Test, that short path root can be read. """
    if conan_version.major == 2:
        return
    new_short_home = Path(tempfile.gettempdir()) / "._myconan_short"
    os.environ["CONAN_USER_HOME_SHORT"] = str(new_short_home)
    if platform.system() == "Windows":
        assert conan_api.get_short_path_root() == new_short_home
    else:
        assert not conan_api.get_short_path_root().exists()
    os.environ.pop("CONAN_USER_HOME_SHORT")



import logging
from abc import abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Sequence, Tuple, Union

from typing_extensions import Self, TypeAlias

from .types import (
    ConanAvailableOptions,
    ConanOptions,
    ConanPackageId,
    ConanPackagePath,
    ConanPkg,
    ConanPkgRef,
    ConanRef,
    ConanSettings,
    EditablePkg,
    Remote,
)

if TYPE_CHECKING:
    from conan_unified_api.cache.conan_cache import ConanInfoCache

UserName: TypeAlias = str
IsAuthenticated: TypeAlias = bool
RemoteName: TypeAlias = str

# Interface and docs


class ConanBaseUnifiedApi:
    """
    API abstraction to provide compatibility between ConanV1 and V2 APIs.
    Thin wrapper:
      If possible, the wrapper will just call the same named method in Conan and add type hints.
      For ConanReferences it will always accept str and ConnReference Objects.
      Return types are almost always objects for References and Packages, so that they can be
      used directly in other functions of this API.

    Paths as return values:
      Paths will set an INVALID_PATH value instead of None if the Path can't be determined.
      This is done, because usually there is always an exists check or something similar,
      which then returns naturally False.
      Otherwise can be checked against the "invalid_path" variable from the conan_unified_api
      namespace.
    Type Hints:
      The wrapper will alias most of the built-in types to the same name and add typedict hints.
      Own implementations are only used to ensure a compatible behavior of conan basic classes
      (like adding the loads method to ConanV2 References)
    Version specific methods:
      In case of a version specific function, the interface will still specify it and the other
      version(s) will simply not do anything or return a default value.
    Exception handling:
      All methods can throw unless noted in the docs. Often the wrapper method will raise a
      ConanException and reference the original error in it.
    """

    @abstractmethod
    def __init__(
        self,
        init: bool = True,
        logger: Optional[logging.Logger] = None,
        mute_logging: bool = False,
    ):
        """
        :param init: Calls init_api function directly in the constructor. Can be disabled for
        faster constructor.
        :param logger: A custom logger can be injected here. Otherwise will use default logger.
        :param mute_logging: Disables the logger, regardless if it was injected ot not.
        Can not raise an exception.
        """
        self.info_cache: ConanInfoCache
        ...

    @abstractmethod
    def init_api(self) -> Self:
        """
        Instantiate the internal Conan api. Can be called extra to split up loading.
        Conan 1 can slow down on init with remove locks noticably, if there are several
        hundred local packages.
        """
        raise NotImplementedError

    @abstractmethod
    def mute_logging(self, mute_logging: bool) -> None:
        """
        Can be used to selectively turn on and off logging for methods.
        :param mute_logging: Disables the logger if set to True, enables it again with false
        """
        raise NotImplementedError

    ### Helper commands ###

    @staticmethod
    @abstractmethod
    def generate_canonical_ref(conan_ref: Union[ConanRef, str]) -> str:
        """Creates a full ref from a short ref, e.g. product/1.0.0 -> product/1.0.0@_/_"""
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def conan_ref_from_reflike(conan_ref: Union[ConanRef, str]) -> ConanRef:
        """Creates a ref object if it is a string otherwise returns the object"""
        raise NotImplementedError

    ### General commands ###

    @abstractmethod
    def info(self, conan_ref: Union[ConanRef, str]) -> List[Dict[str, Any]]:
        """
        Calls the conan info method on Conan V1 to return all recipe and pacakges metainfo
        (including paths) for the recipe itself and all the dependencies. The order is random.
        For the ConanV2 version it calls graph info, which is the best equivalent method.
        Original remote argument ommitted. To download a conanfile from a specific remote
        please use inspect with remote argument.
        """
        raise NotImplementedError

    @abstractmethod
    def inspect(
        self,
        conan_ref: Union[ConanRef, str],
        attributes: Sequence[str] = (),
        remote_name: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Get a field of the selected conanfile. Works currently only with a reference,
        but not with a path. Recreates the ConanV1 behavior for ConanV2.
        If attributes is empty in ConanV1 a chosen default fields are returned, in ConanV2 all
        of them. If attributes is set, only the selected will be returned in the dict.

        """
        raise NotImplementedError

    @abstractmethod
    def alias(
        self,
        conan_ref: Union[ConanRef, str],
        conan_target_ref: Union[ConanRef, str],
    ) -> None:
        """Creates an alias for the target ref with all local packages. For ConanV1 only."""
        raise NotImplementedError

    @abstractmethod
    def remove_locks(self) -> None:
        """Remove local cache locks. For ConanV1 only."""
        raise NotImplementedError

    ### Profile commands ###

    @abstractmethod
    def get_profiles(self) -> List[str]:
        """Return a list of all profiles"""
        raise NotImplementedError

    @abstractmethod
    def get_profile_settings(self, profile_name: str) -> ConanSettings:
        """Return a dict of settings for a profile"""
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def build_conan_profile_name_alias(conan_settings: ConanSettings) -> str:
        """Build a human readable pseduo profile name, like Windows_x64_vs16_v142_release"""
        raise NotImplementedError

    @abstractmethod
    def get_default_settings(self) -> ConanSettings:
        """
        Return the settings of the conan default profile.
        This is not necessarily the same as get_profile_settings("default"),
        because the file content can be changed.
        """
        raise NotImplementedError

    @abstractmethod
    def get_profiles_path(self) -> Path:
        """Get the path to the folder where profiles are located"""
        raise NotImplementedError

    ### Config and settings commands ###

    @abstractmethod
    def get_settings_file_path(self) -> Path:
        """Return conan settings file path (settings.yml)"""
        raise NotImplementedError

    @abstractmethod
    def get_config_file_path(self) -> Path:
        """Return conan config file path (conan.conf)"""
        raise NotImplementedError

    @abstractmethod
    def get_config_entry(self, config_name: str) -> Optional[Any]:
        """Return a conan config entry value (conan.conf).
            Returns None if no entry was set.
        For Conan 1 the format is <section>.<setting_name>, e.g. "general.non_interactive".
            Will evaluate environment variables, which is not consistent to the cli behavior,
            but thus return more accurate results.
            Always returns a string, even for bools.
        For Conan 2 the formt is <section>:<setting_name>, e.g. core:non_interactive
            Returns value correctly typed.
        """
        raise NotImplementedError

    @abstractmethod
    def get_revisions_enabled(self) -> bool:
        """Return if revisions are enabled for Conan V1. Always true in V2 mode."""
        raise NotImplementedError

    @abstractmethod
    def get_user_home_path(self) -> Path:
        """Return Conan user home path, where e.g. settings reside"""
        raise NotImplementedError

    @abstractmethod
    def get_storage_path(self) -> Path:
        """Return Conan storage path where packages are saved"""
        raise NotImplementedError

    @abstractmethod
    def get_short_path_root(self) -> Path:
        """Return short path root for Windows for Conan V1.
        Sadly there is no built-in way to do so."""
        raise NotImplementedError

    ### Remotes ###

    @abstractmethod
    def get_remotes(self, include_disabled: bool = False) -> List[Remote]:
        """Return a list of all remotes as objects"""
        raise NotImplementedError

    @abstractmethod
    def get_remote_names(self, include_disabled: bool = False) -> List[str]:
        """Return a list of all remote names"""
        raise NotImplementedError

    @abstractmethod
    def get_remote(self, remote_name: str) -> Optional[Remote]:
        """Get info of one remote. Returns None on not existing remote."""
        raise NotImplementedError

    @abstractmethod
    def add_remote(self, remote_name: str, url: str, verify_ssl: bool) -> None:
        """Add a new remote with the selected options. Disabled is always false."""
        raise NotImplementedError

    @abstractmethod
    def rename_remote(self, remote_name: str, new_name: str) -> None:
        """Rename a remote"""
        raise NotImplementedError

    @abstractmethod
    def remove_remote(self, remote_name: str) -> None:
        """Remove a remote"""
        raise NotImplementedError

    @abstractmethod
    def update_remote(
        self,
        remote_name: str,
        url: str,
        verify_ssl: bool,
        index: Optional[int] = None,
    ) -> None:
        """Update a remote with new information and reorder with index.
        The Remote object is immutable, so we have to get every field separatly.
        The name can't be changed with this, only with rename.
        """
        raise NotImplementedError

    @abstractmethod
    def disable_remote(self, remote_name: str, disabled: bool) -> None:
        """
        Setting disabled to true disables the remote, settings to false enables it.
        """
        raise NotImplementedError

    @abstractmethod
    def login_remote(self, remote_name: str, user_name: str, password: str) -> None:
        """Login to a remote with credentials
        This method will throw if wrong credentials are entered,
        so we can catch the first error, when logging into multiple remotes
        and do not retry, possibly locking the user.
        """
        raise NotImplementedError

    @abstractmethod
    def get_remote_user_info(self, remote_name: str) -> Tuple[UserName, IsAuthenticated]:
        """Get username and authenticated info for a remote."""
        raise NotImplementedError

    ### Install related methods ###

    @abstractmethod
    def install_reference(
        self,
        conan_ref: Union[ConanRef, str],
        conan_settings: Optional[ConanSettings] = None,
        conan_options: Optional[ConanOptions] = None,
        profile: str = "",
        update: bool = True,
        generators: Sequence[str] = (),
        remote_name: Optional[str] = None,
    ) -> Tuple[ConanPackageId, ConanPackagePath]:
        """
        Try to install a conan reference (without id) with the provided extra information.
        Uses plain conan install (No auto determination of best matching package)
        Returns the actual pkg_id and the package path.
        """
        raise NotImplementedError

    @abstractmethod
    def install_package(
        self,
        conan_ref: Union[ConanRef, str],
        package: ConanPkg,
        update: bool = True,
        remote_name: Optional[str] = None,
    ) -> Tuple[ConanPackageId, ConanPackagePath]:
        """
        Try to install a conan package (id) with the provided extra information.
        Returns the installed id and a valid pkg path, if installation was succesfull.
        WARNING: The installed id can differ from the requested one,
        because there is no built-in way in conan to install a specific package id!
        """
        raise NotImplementedError

    @abstractmethod
    def get_path_with_auto_install(
        self,
        conan_ref: Union[ConanRef, str],
        conan_options: Optional[ConanOptions] = None,
        update: bool = False,
    ) -> Tuple[ConanPackageId, ConanPackagePath]:
        """Return the pkg_id and package folder of a conan reference
        and auto-install it with the best matching package, if it is not available"""
        raise NotImplementedError

    @abstractmethod
    def install_best_matching_package(
        self,
        conan_ref: Union[ConanRef, str],
        conan_options: Optional[ConanOptions] = None,
        update: bool = False,
    ) -> Tuple[ConanPackageId, ConanPackagePath]:
        raise NotImplementedError

    @abstractmethod
    def get_options_with_default_values(
        self,
        conan_ref: Union[ConanRef, str],
        remote_name: Optional[str] = None,
    ) -> Tuple[ConanAvailableOptions, ConanOptions]:
        """Return the available options and their default values as dict."""
        raise NotImplementedError

    ### Local References and Packages ###

    @abstractmethod
    def get_export_folder(self, conan_ref: Union[ConanRef, str]) -> Path:
        """Get the export folder form a reference"""
        raise NotImplementedError

    @abstractmethod
    def get_conanfile_path(self, conan_ref: Union[ConanRef, str]) -> Path:
        """Get local conanfile path. If it is not localy available, download it."""
        raise NotImplementedError

    @abstractmethod
    def get_package_folder(
        self,
        conan_ref: Union[ConanRef, str],
        package_id: str,
    ) -> ConanPackagePath:
        "Get the fully resolved pkg path from the ref and the specific package (id)"
        raise NotImplementedError

    @abstractmethod
    def get_conan_buildinfo(
        self,
        conan_ref: Union[ConanRef, str],
        conan_settings: ConanSettings,
        conan_options: Optional[ConanOptions] = None,
    ) -> str:
        """Read conan buildinfo and return as string
        NOTE: Currently not implemented for Conan 2
        """
        raise NotImplementedError

    @abstractmethod
    def remove_reference(
        self,
        conan_ref: Union[ConanRef, str],
        pkg_id: str = "",
    ) -> None:
        """Remove a conan reference and it's package if specified via id"""
        raise NotImplementedError

    @abstractmethod
    def find_best_matching_local_package(
        self,
        conan_ref: Union[ConanRef, str],
        conan_options: Optional[ConanOptions] = None,
    ) -> ConanPkg:
        """Find a package in the local cache"""
        raise NotImplementedError

    @abstractmethod
    def get_best_matching_local_package_path(
        self,
        conan_ref: Union[ConanRef, str],
        conan_options: Optional[ConanOptions] = None,
    ) -> Tuple[ConanPackageId, ConanPackagePath]:
        """Return the pkg_id and pkg folder of a conan reference, if it is installed."""
        raise NotImplementedError

    @abstractmethod
    def get_all_local_refs(self) -> List[ConanRef]:
        """Returns all locally installed conan references"""
        raise NotImplementedError

    @abstractmethod
    def get_local_pkgs_from_ref(self, conan_ref: Union[ConanRef, str]) -> List[ConanPkg]:
        """Returns all installed pkg ids for a reference."""
        raise NotImplementedError

    @abstractmethod
    def get_local_pkg_from_id(self, pkg_ref: ConanPkgRef) -> ConanPkg:
        """Returns an installed pkg from reference and id"""
        raise NotImplementedError

    @abstractmethod
    def get_local_pkg_from_path(
        self,
        conan_ref: Union[ConanRef, str],
        path: Path,
    ) -> Optional[ConanPkg]:
        """For reverse lookup - give info from path"""
        raise NotImplementedError

    ### Editables ###

    @abstractmethod
    def get_editable(self, conan_ref: Union[ConanRef, str]) -> Optional[EditablePkg]:
        """Get an editable object from conan reference.
        Returns None if editable does not exist.
        """
        raise NotImplementedError

    @abstractmethod
    def get_editables_package_path(self, conan_ref: Union[ConanRef, str]) -> Path:
        """Get package path of an editable reference."""
        raise NotImplementedError

    @abstractmethod
    def get_editables_output_folder(self, conan_ref: Union[ConanRef, str]) -> Optional[Path]:
        """Get output folder of an editable reference."""
        raise NotImplementedError

    @abstractmethod
    def get_editable_references(self) -> List[ConanRef]:
        """Get all local editable references."""
        raise NotImplementedError

    @abstractmethod
    def add_editable(
        self,
        conan_ref: Union[ConanRef, str],
        path: Union[Path, str],
        output_folder: Union[Path, str],
    ) -> bool:
        """Add an editable reference."""
        raise NotImplementedError

    @abstractmethod
    def remove_editable(self, conan_ref: Union[ConanRef, str]) -> bool:
        """Remove an editable reference."""
        raise NotImplementedError

    ### Remote References and Packages ###

    @abstractmethod
    def search_recipes_in_remotes(self, query: str, remote_name: str = "all") -> List[ConanRef]:
        """Search in all remotes for a specific query.
        Returns a list of unique and ordered ConanRefs.
        Limitation: Can't handle long canonical refs like mylib/1.0.0@_/_.
        Must use mylib/1.0.0.
        """
        raise NotImplementedError

    @abstractmethod
    def search_recipe_all_versions_in_remotes(
        self,
        conan_ref: Union[ConanRef, str],
    ) -> List[ConanRef]:
        """Search in all remotes for all versions of a conan ref"""
        raise NotImplementedError

    @abstractmethod
    def get_remote_pkgs_from_ref(
        self,
        conan_ref: Union[ConanRef, str],
        remote_name: Optional[str],
        query: Optional[str] = None,
    ) -> List[ConanPkg]:
        """
        Return all packages for a reference in a specific remote with an optional query.
        Can not raise an exception. Returns an empty list if something errors.
        """
        raise NotImplementedError

    @abstractmethod
    def get_remote_pkg_from_id(self, pkg_ref: ConanPkgRef) -> ConanPkg:
        """Returns a remote pkg from reference and id"""
        raise NotImplementedError

    @abstractmethod
    def find_best_matching_package_in_remotes(
        self,
        conan_ref: Union[ConanRef, str],
        conan_options: Optional[ConanOptions] = None,
    ) -> Tuple[List[ConanPkg], RemoteName]:
        """Find a package with options in the remotes"""
        raise NotImplementedError

    @abstractmethod
    def find_best_matching_packages(
        self,
        conan_ref: Union[ConanRef, str],
        conan_options: Optional[ConanOptions] = None,
        remote_name: Optional[str] = None,
    ) -> List[ConanPkg]:
        """
        Tries to find the best matching packages either locally or in a remote,
        based on the users machine and the supplied options.
        """
        raise NotImplementedError



import json
from multiprocessing import RLock
from pathlib import Path
from typing import Dict, List, Optional, Set, Union

from conan_unified_api.base import conan_version
from conan_unified_api.base.helper import delete_path
from conan_unified_api.base.logger import Logger
from conan_unified_api.common import ConanUnifiedApi
from conan_unified_api.types import ConanRef


class ConanInfoCache:
    """
    This is a cache to accelerate calls which need remote access.
    It also has an option to store the local package path.
    """

    CACHE_FILE_NAME = "cache.json"
    if conan_version.major == 2:
        CACHE_FILE_NAME = "cacheV2.json"

    def __init__(self, cache_dir: Path, local_refs: Optional[List[ConanRef]] = None):
        if not local_refs:
            local_refs = []

        self._cache_file = cache_dir / self.CACHE_FILE_NAME
        self._remote_packages: Dict[str, Dict[str, List[str]]] = {}
        self._read_only = False  # for testing purposes
        self._access_lock = RLock()

        # create cache file, if it does not exist
        if not self._cache_file.exists():
            self._cache_file.touch()
            return

        # read cached info
        self._load()

    def get_similar_remote_pkg_refs(self, name: str, user: str) -> List[ConanRef]:
        """
        Return cached info on remotely available conan refs from the same ref name and user.
        """
        if not user:  # official pkgs have no user, substituted by _
            user = "_"
        refs: List[ConanRef] = []
        with self._access_lock:
            if user == "*":  # find all refs with same name
                name_pkgs = self._remote_packages.get(name, {})
                for user in name_pkgs:
                    for version_channel in self._remote_packages.get(name, {}).get(user, []):
                        version, channel = version_channel.split("/")
                        refs.append(ConanRef(name, version, user, channel))
            else:
                user_pkgs = self._remote_packages.get(name, {}).get(user, [])
                for version_channel in user_pkgs:
                    version, channel = version_channel.split("/")
                    refs.append(ConanRef(name, version, user, channel))
        return refs

    def get_all_remote_refs(self) -> List[str]:
        """Return all remote references. Updating, when queries finish."""
        refs = []
        with self._access_lock:
            for name in self._remote_packages:
                for user in self._remote_packages[name]:
                    for version_channel in self._remote_packages.get(name, {}).get(user, []):
                        version, channel = version_channel.split("/")
                        refs.append(str(ConanRef(name, version, user, channel)))
        return refs

    def search(self, query: str) -> Set[str]:
        """
        Return cached info on available conan refs from a query
        """
        with self._access_lock:
            remote_refs = set()
            # try to extract name and user from query
            split_query = query.split("/")
            name = split_query[0]
            user = "*"
            if len(split_query) > 1:
                user_split = split_query[1].split("@")
                if len(user_split) > 1:
                    user = user_split[1]
            for ref in self.get_similar_remote_pkg_refs(name, user):
                remote_refs.add(str(ref))
            return remote_refs

    def invalidate_remote_package(self, conan_ref: Union[ConanRef, str]) -> None:
        """Remove a package, wich was removed on the remote"""
        conan_ref = ConanUnifiedApi.conan_ref_from_reflike(conan_ref)
        version_channels = self._remote_packages.get(conan_ref.name, {}).get(
            str(conan_ref.user), []
        )
        invalid_version_channel = f"{conan_ref.version}/{conan_ref.channel}"
        if invalid_version_channel in version_channels:
            Logger().debug(f"Invalidated {conan_ref!s} from remote cache.")
            version_channels.remove(f"{conan_ref.version}/{conan_ref.channel}")
            self._save()

    def update_remote_package_list(
        self, remote_packages: List[ConanRef], invalidate: bool = False
    ) -> None:
        """
        Update the cache with the info of several remote packages.
        Invalidate option clears the cache.
        """
        with self._access_lock:
            if invalidate:  # clear cache
                self._remote_packages.clear()
                self._save()
            for ref in remote_packages:
                # convert back the official cache entries
                user = str(ref.user)
                channel = ref.channel
                if ref.user is None and ref.channel is None:
                    user = "_"
                    channel = "_"
                current_version_channel = f"{ref.version}/{channel}"
                version_channels = set(self._remote_packages.get(ref.name, {}).get(user, []))
                if current_version_channel not in version_channels:
                    version_channels.add(current_version_channel)
                    version_channels_list = list(version_channels)
                    if not self._remote_packages.get(ref.name):
                        self._remote_packages.update({ref.name: {}})
                    self._remote_packages.get(ref.name, {}).update(
                        {user: version_channels_list}
                    )

            self._save()

    def _load(self) -> None:
        """Load the cache."""
        json_data = {}
        try:
            with open(self._cache_file) as json_file:
                content = json_file.read()
                if len(content) > 0:
                    json_data = json.loads(content)
        except Exception:  # possibly corrupt, delete cache file
            Logger().debug("ConanCache: Can't read speedup-cache file, deleting it.")
            delete_path(self._cache_file)
            # create file anew
            self._cache_file.touch()
            return
        self._remote_packages = json_data.get("remote_packages", {})
        self._read_only = json_data.get("read_only", False)

    def _save(self) -> None:
        """Write the cache to file."""
        if self._read_only:
            return
        json_data = {}
        with self._access_lock:
            json_data["read_only"] = self._read_only
            json_data["remote_packages"] = self._remote_packages
            try:
                with open(self._cache_file, "w") as json_file:
                    json.dump(json_data, json_file)
            except Exception:
                Logger().debug("ConanCache: Can't save speedup-cache file.")



"""OS Abstraction Layer for all file based functions"""

import os
import stat
import sys
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Callable, Dict, List

from conan_unified_api.base.logger import Logger


def str2bool(value: str) -> bool:
    """Own impl. isntead of distutils.util.strtobool
    because distutils will be deprecated"""
    value = value.lower()
    if value in {"yes", "true", "y", "1"}:
        return True
    if value in {"no", "false", "n", "0"}:
        return False
    return False


def create_key_value_pair_list(input_dict: Dict[str, Any]) -> List[str]:
    """
    Helper to create name=value string list from dict
    Filters "ANY" options.
    """
    res_list: List[str] = []
    if not input_dict:
        return res_list
    for name, value in input_dict.items():
        value_str = str(value)
        # this is not really safe, but there can be wild values...
        if "any" in value_str.lower() or "none" in value_str.lower():
            continue
        res_list.append(name + "=" + value_str)
    return res_list


def delete_path(dst: Path) -> None:
    """
    Delete file or (non-empty) folder recursively.
    Exceptions will be caught and message logged to stdout.
    """
    from shutil import rmtree

    try:
        if dst.is_file():
            dst.unlink()
        elif dst.is_dir():

            def rm_dir_readonly(func: Callable, path: str, _) -> None:
                "Clear the readonly bit and reattempt the removal"
                os.chmod(path, stat.S_IWRITE)  # noqa: PTH101
                func(path)

            rmtree(str(dst), onerror=rm_dir_readonly)  # onexc only usable from 3.12
    except Exception as e:
        Logger().warning("Can't delete %s: %s", str(dst), str(e))


@contextmanager
def save_sys_path():  # noqa: ANN201
    saved_path = sys.path.copy()
    yield
    # restore
    sys.path = saved_path



import os
import tempfile
from pathlib import Path


from conan_unified_api.base.helper import (create_key_value_pair_list, delete_path)
from conan_unified_api.unified_api import ConanBaseUnifiedApi
from conan_unified_api.common import ConanUnifiedApi
from test import TEST_REF_OFFICIAL


def test_delete():
    """ 
    1. Delete file
    2. Delete non-empty directory
    """
    # 1. Delete file
    test_file = Path(tempfile.mkdtemp()) / "test.inf"
    test_file_content = "test"
    with open(str(test_file), "w") as f:
        f.write(test_file_content)
    delete_path(test_file)
    assert not test_file.exists()

    # 2. Delete non-empty directory
    test_dir = Path(tempfile.mkdtemp()) / "test_dir"
    os.makedirs(test_dir)
    test_dir_file = test_dir / "test.inf"
    with open(str(test_dir_file), "w") as f:
        f.write("test")
    delete_path(test_dir)
    assert not test_dir.exists()


def test_create_key_value_list():
    """
    Test, that key value pairs can be extracted as strings. No arrays or other tpyes supported.
    The return value must be a list of strings in the format ["key1=value1", "key2=value2]
    "Any" values are ignored. (case insensitive)
    """
    inp = {"Key1": "Value1"}
    res = create_key_value_pair_list(inp)
    assert res == ["Key1=Value1"]
    inp = {"Key1": "Value1", "Key2": "Value2"}
    res = create_key_value_pair_list(inp)
    assert res == ["Key1=Value1", "Key2=Value2"]
    inp = {"Key1": "Value1", "Key2": "Any"}
    res = create_key_value_pair_list(inp)
    assert res == ["Key1=Value1"]

def test_generate_canonical_ref(conan_api: ConanBaseUnifiedApi):
    ref = conan_api.generate_canonical_ref(TEST_REF_OFFICIAL.split("@")[0])
    assert ref == TEST_REF_OFFICIAL


def test_resolve_default_options(conan_api: ConanUnifiedApi):
    """
    Test, if different kind of types of default options can be converted to a dict
    Dict is expected.
    """
    str_val = "option=value"
    ret = conan_api._resolve_default_options(str_val)
    assert ret.items()

    tup_val = ("option=value", "options2=value2")
    ret = conan_api._resolve_default_options(tup_val)
    assert ret.items()

    list_val = ["option=value", "options2=value2"]
    ret = conan_api._resolve_default_options(list_val)
    assert ret.items()



import logging
import os
from contextlib import redirect_stderr, redirect_stdout
from pathlib import Path
from tempfile import gettempdir
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Sequence, Tuple, Union
from unittest.mock import patch

from typing_extensions import Self

from .base import INVALID_PATH_VALUE, Version, conan_version
from .base.helper import create_key_value_pair_list
from .base.logger import Logger
from .base.typing import SignatureCheckMeta
from .common import ConanUnifiedApi
from .types import (
    ConanAvailableOptions,
    ConanException,
    ConanOptions,
    ConanPackageId,
    ConanPackagePath,
    ConanPkg,
    ConanPkgRef,
    ConanRef,
    ConanSettings,
    EditablePkg,
    Remote,
)

current_path = Path(__file__).parent

if TYPE_CHECKING:
    from conan.api.conan_api import ConanAPI  # type: ignore
    from conan.internal.cache.cache import PkgCache as ClientCache
    from conan.internal.cache.home_paths import HomePaths  # type: ignore

    from .cache.conan_cache import ConanInfoCache


class ConanApi(ConanUnifiedApi, metaclass=SignatureCheckMeta):
    """Wrapper around ConanAPIV2"""

    def __init__(
        self,
        init: bool = True,
        logger: Optional[logging.Logger] = None,
        mute_logging: bool = False,
    ):
        self.info_cache: ConanInfoCache
        self._conan: ConanAPI
        self._client_cache: ClientCache
        self._short_path_root = Path("Unknown")
        self._home_paths: HomePaths
        super().__init__(init, logger, mute_logging)

    def init_api(self) -> Self:
        from conan.api.conan_api import ConanAPI

        devnull = open(os.devnull, "w")
        with redirect_stdout(devnull), redirect_stderr(devnull):
            self._conan = ConanAPI()
            self._init_client_cache()
        devnull.close()
        from .cache.conan_cache import ConanInfoCache

        self.info_cache = ConanInfoCache(current_path, self.get_all_local_refs())
        self.logger.debug("Initialized Conan V2 API wrapper")
        return self

    def _init_client_cache(self) -> None:
        if conan_version < Version("2.4.0"):
            from conans.client.cache.cache import ClientCache
        else:
            from conan.internal.cache.cache import PkgCache as ClientCache
        self._client_cache = ClientCache(
            self._conan.cache_folder, self._conan.config.global_conf
        )
        from conan.internal.cache.home_paths import HomePaths

        self._home_paths = HomePaths(self._conan.cache_folder)

    ### General commands ###

    def info(self, conan_ref: Union[ConanRef, str]) -> List[Dict[str, Any]]:
        # TODO: try to merge with install reference code
        conan_ref = self.conan_ref_from_reflike(conan_ref)
        remotes = self._conan.remotes.list(None)
        profiles = []
        profile_host = self._conan.profiles.get_profile(profiles, settings=[], options=[])
        requires = [conan_ref]
        update = "*" if conan_version < Version("2.2") else False

        deps_graph = self._conan.graph.load_graph_requires(
            requires, None, profile_host, profile_host, None, remotes, update
        )
        # this modifies deps_graph
        self._conan.graph.analyze_binaries(
            deps_graph, build_mode=None, remotes=remotes, update=update, lockfile=None
        )

        # prepare output and convert to dict
        nodes = deps_graph.nodes
        nodes.pop(0)  # remove cli node
        results = []
        for node in nodes:
            result_dict = {}
            for attr in dir(node):
                if attr.startswith("_"):  # only public fields
                    continue
                try:
                    result_dict[attr] = getattr(node, attr)
                except Exception:  # almost anything can go wrong here, ignore
                    continue
            results.append(result_dict)
        return results

    def inspect(
        self,
        conan_ref: Union[ConanRef, str],
        attributes: Sequence[str] = (),
        remote_name: Optional[str] = None,
    ) -> Dict[str, Any]:
        conan_ref = self.conan_ref_from_reflike(conan_ref)

        remotes = self.get_remotes()
        if remote_name:
            remotes = [self.get_remote(remote_name)]

        path = self.get_conanfile_path(conan_ref)
        if conan_version < Version("2.4"):
            conanfile = self._conan.local.inspect(str(path), remotes=remotes, lockfile=None)
        else:
            conanfile = self._conan.local.inspect(
                str(path),
                remotes=remotes,
                lockfile=None,
                name=conan_ref.name,
                version=conan_ref.version,
                user=conan_ref.user,
                channel=conan_ref.channel,
            )
        result = {}
        for attr in dir(conanfile):
            if attr.startswith("_"):  # only public fields
                continue
            if attributes and attr not in attributes:
                continue
            try:
                result[attr] = getattr(conanfile, attr)
            except Exception:
                continue
        # no serialization, like in ConanV1
        return result

    # def alias(self, conan_ref: Union[ConanRef, str], conan_target_ref: Union[ConanRef, str]):
    #     # TODO does not seem to copy the package yet
    #     conan_ref = self.conan_ref_from_reflike(conan_ref)
    #     conan_target_ref = self.conan_ref_from_reflike(conan_target_ref)

    #     template = self._conan.new.get_builtin_template("alias")
    #     content = self._conan.new.render(template, {
    #         "name": conan_ref.name,
    #         "version": conan_ref.version,
    #         "target": conan_target_ref.version})
    #     conanfile_path = Path(mkdtemp()) / "conanfile_temp.py"
    #     conanfile_path.write_text(content.get("conanfile.py", ""))
    #     self._conan.export.export(str(conanfile_path), conan_ref.name,
    #                               conan_ref.version, conan_ref.user, conan_ref.channel)

    def remove_locks(self) -> None:
        pass  # command does not exist

    def get_profiles(self) -> List[str]:
        return self._conan.profiles.list()

    def get_profile_settings(self, profile_name: str) -> ConanSettings:
        if conan_version < Version("2.6.0"):
            from conans.client.profile_loader import ProfileLoader
        else:
            from conan.internal.api.profile.profile_loader import ProfileLoader

        try:
            profile = ProfileLoader(self._conan.cache_folder).load_profile(profile_name)
        except Exception as e:
            msg = f"Can't get profile {profile_name} settings: {e!s}"
            raise ConanException(msg) from e
        return profile.settings

    def get_package_folder(self, conan_ref: Union[ConanRef, str], package_id: str) -> Path:
        if not package_id:  # will give the base path ortherwise
            return Path(INVALID_PATH_VALUE)
        conan_ref = self.conan_ref_from_reflike(conan_ref)
        try:
            latest_rev_ref = self._conan.list.latest_recipe_revision(conan_ref)
            latest_rev_pkg = self._conan.list.latest_package_revision(
                ConanPkgRef(latest_rev_ref, package_id)
            )
            layout = self._client_cache.pkg_layout(latest_rev_pkg)  # type: ignore
            return Path(layout.package())
        except Exception:  # gotta catch 'em all!
            return Path(INVALID_PATH_VALUE)

    def get_export_folder(self, conan_ref: Union[ConanRef, str]) -> Path:
        conan_ref = self.conan_ref_from_reflike(conan_ref)
        return Path(self._conan.cache.export_path(conan_ref))

    def get_conanfile_path(self, conan_ref: Union[ConanRef, str]) -> Path:
        conan_ref = self.conan_ref_from_reflike(conan_ref)
        try:
            if conan_ref not in self.get_all_local_refs():
                for remote in self.get_remotes():
                    result = None
                    try:
                        result = self.search_recipes_in_remotes(
                            str(conan_ref), remote_name=remote.name
                        )
                    except Exception:
                        continue  # next
                    if result:
                        latest_rev: ConanRef = self._conan.list.latest_recipe_revision(
                            conan_ref, remote
                        )
                        self._conan.download.recipe(latest_rev, remote)  # type: ignore
                        break
            path = self._conan.local.get_conanfile_path(
                self._conan.cache.export_path(conan_ref), str(Path.cwd()), py=True
            )
            if not path:
                return Path(INVALID_PATH_VALUE)
            return Path(path)
        except Exception as e:
            Logger().error(f"Can't get conanfile: {e!s}")
        return Path(INVALID_PATH_VALUE)

    def get_default_settings(self) -> ConanSettings:
        if conan_version < Version("2.6.0"):
            from conans.client.profile_loader import ProfileLoader
        else:
            from conan.internal.api.profile.profile_loader import ProfileLoader

        profile = ProfileLoader(self._conan.cache_folder).load_profile(
            Path(self._conan.profiles.get_default_host()).name
        )

        return dict(profile.settings)

    # user_name, authenticated
    def get_remote_user_info(self, remote_name: str) -> Tuple[str, bool]:
        try:
            info = self._conan.remotes.user_info(self._conan.remotes.get(remote_name))
        except Exception as e:
            msg = f"Can't get remote {remote_name} user info: {e!s}"
            raise ConanException(msg) from e
        return str(info.get("user_name", "")), info.get("authenticated", False)

    def get_config_file_path(self) -> Path:
        return Path(self._home_paths.global_conf_path)

    def get_config_entry(self, config_name: str) -> Optional[Any]:
        return self._conan.config.get(config_name, None)

    def get_revisions_enabled(self) -> bool:
        return True  # always on in 2

    def get_settings_file_path(self) -> Path:
        return Path(self._home_paths.settings_path)

    def get_profiles_path(self) -> Path:
        return Path(self._home_paths.profiles_path)

    def get_user_home_path(self) -> Path:
        return Path(self._conan.cache_folder)

    def get_storage_path(self) -> Path:
        return Path(str(self._client_cache.store))

    def get_short_path_root(self) -> Path:
        # there is no short_paths feature in conan 2
        return Path(INVALID_PATH_VALUE)

    # Remotes

    def get_remotes(self, include_disabled: bool = False) -> List[Remote]:
        remotes = []
        try:
            remotes = self._conan.remotes.list(None, only_enabled=not include_disabled)
        except Exception as e:
            msg = f"Error while reading remotes: {e!s}"
            raise ConanException(msg) from e
        return remotes  # type: ignore

    def add_remote(self, remote_name: str, url: str, verify_ssl: bool) -> None:
        if conan_version < Version("2.1"):
            from conans.client.cache.remote_registry import Remote as ConanRemote
        else:
            from conan.api.model import Remote as ConanRemote
        remote = ConanRemote(remote_name, url, verify_ssl, False)
        self._conan.remotes.add(remote)

    def rename_remote(self, remote_name: str, new_name: str) -> None:
        self._conan.remotes.rename(remote_name, new_name)

    def remove_remote(self, remote_name: str) -> None:
        self._conan.remotes.remove(remote_name)

    def disable_remote(self, remote_name: str, disabled: bool) -> None:
        if disabled:
            self._conan.remotes.disable(remote_name)
        else:
            self._conan.remotes.enable(remote_name)

    def update_remote(
        self, remote_name: str, url: str, verify_ssl: bool, index: Optional[int] = None
    ) -> None:
        self._conan.remotes.update(
            remote_name,
            url,
            verify_ssl,
            self._conan.remotes.get(remote_name).disabled,
            index,
        )

    def login_remote(self, remote_name: str, user_name: str, password: str) -> None:
        if conan_version < Version("2.1"):
            self._conan.remotes.login(self._conan.remotes.get(remote_name), user_name, password)
        else:
            self._conan.remotes.user_login(
                self._conan.remotes.get(remote_name), user_name, password
            )

    ### Install related methods ###
    def install_reference(
        self,
        conan_ref: Union[ConanRef, str],
        conan_settings: Optional[ConanSettings] = None,
        conan_options: Optional[ConanOptions] = None,
        profile: str = "",
        update: bool = True,
        generators: Sequence[str] = (),
        remote_name: Optional[str] = None,
    ) -> Tuple[ConanPackageId, ConanPackagePath]:
        conan_ref = self.conan_ref_from_reflike(conan_ref)
        pkg_id = ""
        if conan_options is None:
            conan_options = {}
        if conan_settings is None:
            conan_settings = {}
        options_list = create_key_value_pair_list(conan_options)
        settings_list = create_key_value_pair_list(conan_settings)
        install_message = (
            f"Installing '{conan_ref!s}' with profile: {profile}, "
            f"settings: {settings_list!s}, "
            f"options: {options_list!s} and update={update}\n"
        )
        Logger().info(install_message)
        from conan.cli.printers.graph import print_graph_basic, print_graph_packages

        try:
            # Basic collaborators, remotes, lockfile, profiles
            remotes = self._conan.remotes.list(None)
            if remote_name:
                remotes = [self.get_remote(remote_name)]
            profiles = [profile] if profile else []
            profile_host = self._conan.profiles.get_profile(
                profiles, settings=settings_list, options=options_list
            )
            requires = [conan_ref]
            if conan_version > Version("2.0"):
                update = "*" if update else ""
            deps_graph = self._conan.graph.load_graph_requires(
                requires, None, profile_host, profile_host, None, remotes, update
            )
            print_graph_basic(deps_graph)
            deps_graph.report_graph_error()
            self._conan.graph.analyze_binaries(
                deps_graph,
                build_mode=None,
                remotes=remotes,
                update=update,
                lockfile=None,
            )
            print_graph_packages(deps_graph)

            # Try to redirect custom streams in conanfile, to avoid missing flush method
            devnull = open(os.devnull, "w")
            # also spoof os.terminal_size(
            spoof_size = os.terminal_size([80, 20])
            patched_tersize = patch("os.get_terminal_size")
            with redirect_stdout(devnull), redirect_stderr(devnull):
                mock = patched_tersize.start()
                mock.return_value = spoof_size

                self._conan.install.install_binaries(deps_graph=deps_graph, remotes=remotes)
                # Currently unused
                self._conan.install.install_consumer(
                    deps_graph=deps_graph,
                    generators=generators,
                    output_folder=None,
                    source_folder=gettempdir(),
                )

                patched_tersize.stop()

            devnull.close()

            info = None
            for node in deps_graph.nodes:
                if node.ref == conan_ref:
                    info = node
                    break
            if info is None:
                msg = "Can't read information of installed recipe from graph."
                raise ConanException(msg)
            pkg_id = info.package_id
            Logger().info(f"Installation of '{conan_ref!s}' finished")
            return (pkg_id, self.get_package_folder(conan_ref, pkg_id))
        except ConanException as e:
            msg = f"Can't install reference '{conan_ref!s}': {e!s}"
            raise ConanException(msg) from e

    def get_options_with_default_values(
        self, conan_ref: Union[ConanRef, str], remote_name: Optional[str] = None
    ) -> Tuple[ConanAvailableOptions, ConanOptions]:
        # this calls external code of the recipe
        conan_ref = self.conan_ref_from_reflike(conan_ref)
        default_options: ConanOptions = {}
        available_options: ConanAvailableOptions = {}
        try:
            self.inspect(conan_ref, remote_name=remote_name)
            path = self.get_conanfile_path(conan_ref)
            from conan.internal.conan_app import ConanApp

            if conan_version < Version("2.1"):
                app = ConanApp(self._conan.cache_folder, self._conan.config.global_conf)
            else:
                app = ConanApp(self._conan)
            conanfile = app.loader.load_conanfile(path, conan_ref)
            default_options = {}
            if conanfile.default_options is not None:
                default_options = conanfile.default_options
            default_options = self._resolve_default_options(default_options)
            available_options = conanfile.options.possible_values
        except Exception as e:  # silent error - if we have no options don't spam the user
            Logger().debug(f"Error while getting default options for {conan_ref!s}: {e!s}")
        return available_options, default_options

    ### Local References and Packages ###

    def get_conan_buildinfo(
        self,
        conan_ref: Union[ConanRef, str],
        conan_settings: ConanSettings,
        conan_options: Optional[ConanOptions] = None,
    ) -> str:
        """TODO: Currently there is no equivalent to txt generator from ConanV1"""
        raise NotImplementedError

    def get_editable(self, conan_ref: Union[ConanRef, str]) -> Optional[EditablePkg]:
        conan_ref = self.conan_ref_from_reflike(conan_ref)
        self.__reload_conan_api()
        editables_dict = self._conan.local.editable_list()
        editable_dict = editables_dict.get(conan_ref, {})
        if not editable_dict:
            return None
        return EditablePkg(
            str(conan_ref),
            editable_dict.get("path", INVALID_PATH_VALUE),
            editable_dict.get("output_folder"),
        )

    def get_editables_package_path(self, conan_ref: Union[ConanRef, str]) -> Path:
        """Get package path of an editable reference. Can be a folder or conanfile.py"""
        conan_ref = self.conan_ref_from_reflike(conan_ref)
        self.__reload_conan_api()
        editables_dict = self._conan.local.editable_list()
        return Path(editables_dict.get(conan_ref, {}).get("path", INVALID_PATH_VALUE))

    def get_editables_output_folder(self, conan_ref: Union[ConanRef, str]) -> Optional[Path]:
        conan_ref = self.conan_ref_from_reflike(conan_ref)
        self.__reload_conan_api()
        editables_dict = self._conan.local.editable_list()
        output_folder = editables_dict.get(conan_ref, {}).get("output_folder")
        if not output_folder:
            return None
        return Path(str(output_folder))

    def get_editable_references(self) -> List[ConanRef]:
        """Get all local editable references."""
        self.__reload_conan_api()
        editables_dict = self._conan.local.editable_list()
        return list(editables_dict.keys())

    def __reload_conan_api(self) -> None:
        from conan.api.conan_api import ConanAPI

        self._conan = ConanAPI()  # reload editables only possible like this

    def add_editable(
        self,
        conan_ref: Union[ConanRef, str],
        path: Union[Path, str],
        output_folder: Union[Path, str],
    ) -> bool:
        try:
            conan_ref = self.conan_ref_from_reflike(conan_ref)
            self._conan.local.editable_add(
                str(path),
                conan_ref.name,
                conan_ref.version,
                conan_ref.user,
                conan_ref.channel,
                output_folder=str(output_folder),
            )
        except Exception as e:
            self.logger.error("Error adding editable: %s", str(e))
            return False
        return True

    def remove_editable(self, conan_ref: Union[ConanRef, str]) -> bool:
        try:
            conan_ref = self.conan_ref_from_reflike(conan_ref)
            self._conan.local.editable_remove(None, [str(conan_ref)])
        except Exception as e:
            self.logger.error("Error removing editable: %s", str(e))
            return False
        return True

    def remove_reference(self, conan_ref: Union[ConanRef, str], pkg_id: str = "") -> None:
        if pkg_id:
            conan_pkg_ref = ConanPkgRef.loads(str(conan_ref) + ":" + pkg_id)
            latest_rev = self._conan.list.latest_package_revision(conan_pkg_ref)
            self._conan.remove.package(latest_rev, remote=None)  # type: ignore
        else:
            conan_ref = self.conan_ref_from_reflike(conan_ref)
            latest_rev = self._conan.list.latest_recipe_revision(conan_ref)
            self._conan.remove.recipe(latest_rev, remote=None)  # type: ignore

    def get_all_local_refs(self) -> List[ConanRef]:
        if conan_version < Version("2.6"):
            return self._client_cache.all_refs()
        return self._client_cache._db.list_references()  # noqa: SLF001

    def get_local_pkgs_from_ref(self, conan_ref: Union[ConanRef, str]) -> List[ConanPkg]:
        conan_ref = self.conan_ref_from_reflike(conan_ref)
        result: List[ConanPkg] = []

        if conan_ref.user == "_":
            conan_ref.user = None
        if conan_ref.channel == "_":
            conan_ref.channel = None
        if not conan_ref.revision:
            try:
                conan_ref_latest: ConanRef = self._conan.list.latest_recipe_revision(conan_ref)
            except Exception as e:
                Logger().error(f"Error while getting latest recipe for {conan_ref!s}: {e!s}")
                return result
            if not conan_ref_latest:
                return result
        else:
            conan_ref_latest = conan_ref
        try:  # errors with invalid pkg
            refs = self._conan.list.packages_configurations(conan_ref_latest)
        except Exception as e:
            Logger().error(f"Error while getting packages for recipe {conan_ref!s}: {e!s}")
            return result
        for ref, pkg_info in refs.items():
            pkg = ConanPkg()
            pkg["id"] = str(ref.package_id)
            pkg["options"] = pkg_info.get("options", {})
            pkg["settings"] = pkg_info.get("settings", {})
            pkg["requires"] = []
            pkg["outdated"] = False
            result.append(pkg)
        return result

    ### Remote References and Packages ###

    def search_recipes_in_remotes(self, query: str, remote_name: str = "all") -> List[ConanRef]:
        search_results = []
        remotes = []
        if remote_name == "all":
            remotes = self.get_remotes()
        else:
            remotes = [self.get_remote(remote_name)]
            if not remotes:
                msg = f"Error while searching for recipe: remote {remote_name} does not exist"
                raise ConanException(msg)
        for remote in remotes:
            try:
                # no query possible with pattern
                search_results: List[ConanRef] = self._conan.search.recipes(
                    query, remote=remote
                )
            except Exception as e:
                Logger().error(f"Error while searching for recipe: {e!s}")

        search_results = list(set(search_results))  # make unique
        search_results.sort()
        return search_results

    def search_recipe_all_versions_in_remotes(
        self, conan_ref: Union[ConanRef, str]
    ) -> List[ConanRef]:
        conan_ref = self.conan_ref_from_reflike(conan_ref)
        search_results = []
        search_results: List = self.search_recipes_in_remotes(
            f"{conan_ref.name}/*", remote_name="all"
        )

        self.info_cache.update_remote_package_list(search_results)
        return search_results

    def get_remote_pkgs_from_ref(
        self,
        conan_ref: Union[ConanRef, str],
        remote_name: Optional[str],
        query: Optional[str] = None,
    ) -> List[ConanPkg]:
        conan_ref = self.conan_ref_from_reflike(conan_ref)

        found_pkgs: List[ConanPkg] = []
        try:
            from conan.api.model import ListPattern

            pattern = ListPattern(str(conan_ref) + ":*", rrev="", prev="")
            search_results = None
            remote_obj = None
            if remote_name:
                for remote_obj in self.get_remotes():
                    if remote_obj.name == remote_name:
                        break
            search_results = self._conan.list.select(
                pattern, remote=remote_obj, package_query=query
            )
            if search_results:
                latest_rev = self._conan.list.latest_recipe_revision(conan_ref, remote_obj)
                if latest_rev:
                    found_pkgs_dict = (
                        search_results.recipes.get(str(conan_ref), {})
                        .get("revisions", {})
                        .get(latest_rev.revision, {})
                        .get("packages", {})
                    )
                    for id, info in found_pkgs_dict.items():
                        found_pkgs.append(
                            ConanPkg(
                                id=id,
                                options=info.get("info", {}).get("options", {}),
                                settings=info.get("info", {}).get("settings", {}),
                                requires=[],
                                outdated=False,
                            )
                        )
            Logger().debug(str(found_pkgs))
        except ConanException as e:
            error_message = f"Can not get Conan packages for reference {conan_ref}: {e!s}"
            Logger().error(str(error_message))
        return found_pkgs



"""
Convenience access to all user related classes and factory for conan api.
"""

import logging
from importlib.metadata import distribution
from pathlib import Path
from typing import Optional

from .base import conan_version, invalid_path
from .cache.conan_cache import ConanInfoCache
from .common import ConanUnifiedApi

PKG_NAME = "conan_unified_api"
__version__ = distribution(PKG_NAME).version

# Paths to find folders - points to the folder of this file
# must be initialized later, otherwise setup.py can't parse this file

base_path = Path(__file__).absolute().parent


def ConanApiFactory(  # noqa: N802
    init: bool = True,
    logger: Optional[logging.Logger] = None,
    mute_logging: bool = False,
) -> ConanUnifiedApi:
    """Instantiate ConanApi in the correct version"""
    if conan_version.major == 1:
        from conan_unified_api.conan_v1 import ConanApi
    elif conan_version.major == 2:
        from .conan_v2 import ConanApi
    else:
        msg = "Can't recognize Conan version"
        raise RuntimeError(msg)

    return ConanApi(init, logger, mute_logging)


__all__ = [
    "ConanInfoCache",
    "base_path",
    "invalid_path",
]




import pytest
from pytest_check import check
from conan_unified_api.unified_api import ConanBaseUnifiedApi
from test.conan_helper import disable_remote, remove_remote, add_remote, TEST_REMOTE_NAME
from test import TEST_REMOTE_URL, TEST_REMOTE_USER, time_function

@pytest.fixture
def new_remote():
    """ Fixture factory for multiple new remotes. Cleans up after each testcase. """
    remotes = []

    def _add_remote(name="new1", url="http://localhost:9303"):
        remove_remote(name)
        add_remote(name, url)
        remotes.append(name)
        return name, url

    yield _add_remote

    for remote in remotes:
        remove_remote(remote)


def test_add_remove_remotes(conan_api: ConanBaseUnifiedApi):
    """ Check that adding a new remote adds it with all used options.
    Afterwards delete it and check.
    """
    test_remote_name = "new1"
    # remove with cli to ensure that new1 does not exist
    with time_function("remove_remote_setup"):
        remove_remote(test_remote_name)

    orig_remotes = conan_api.get_remotes()
    with time_function("add_remote"):
        conan_api.add_remote(test_remote_name, "http://localhost:9301", False)

    new_remote = conan_api.get_remotes()[-1]
    assert new_remote.name == test_remote_name
    assert new_remote.url == "http://localhost:9301"
    assert not new_remote.verify_ssl

    with time_function("remove_remote"):
        conan_api.remove_remote(test_remote_name)

    assert len(conan_api.get_remotes()) == len(orig_remotes)


def test_disable_remotes(conan_api: ConanBaseUnifiedApi, new_remote):
    new_remote_name, _ = new_remote()
    remote = conan_api.get_remote(new_remote_name)
    assert remote
    assert not remote.disabled

    conan_api.disable_remote(new_remote_name, True)
    remote = conan_api.get_remote(new_remote_name)
    assert remote
    assert remote.disabled

    conan_api.disable_remote(new_remote_name, False)
    remote = conan_api.get_remote(new_remote_name)
    assert remote
    assert not remote.disabled

def test_get_remote_user_info(conan_api: ConanBaseUnifiedApi):
    """ Check that get_remote_user_info returns a tuple of name and login
      state for the test remote """
    info = conan_api.get_remote_user_info(TEST_REMOTE_NAME)
    assert info == (TEST_REMOTE_USER, True)


def test_get_remotes(conan_api: ConanBaseUnifiedApi, new_remote):
    """ Test that get_remotes returns remote objects and cotains the test remote and 
    the new remote. Also check include_disabled flag.
    """
    new_remote_name, _ = new_remote()

    remotes = conan_api.get_remotes()
    assert len(remotes) >= 2
    found_remote = 0
    for remote in remotes:
        if remote.name ==  TEST_REMOTE_NAME:
            found_remote += 1
        if remote.name == new_remote_name:
            found_remote += 1
    assert found_remote == 2
    disable_remote(new_remote_name)
    remotes = conan_api.get_remotes(include_disabled=True)
    assert remotes[-1].name == new_remote_name


def test_get_remotes_names(conan_api: ConanBaseUnifiedApi, new_remote):
    new_remote_name, _ = new_remote()

    disable_remote(new_remote_name)

    remote_names = conan_api.get_remote_names()
    assert TEST_REMOTE_NAME in remote_names
    assert new_remote_name not in remote_names

    remote_names = conan_api.get_remote_names(include_disabled=True)
    assert TEST_REMOTE_NAME in remote_names
    assert new_remote_name in remote_names


def test_get_remote(conan_api: ConanBaseUnifiedApi):
    remote = conan_api.get_remote(TEST_REMOTE_NAME)
    assert remote  # not None
    assert remote.name == TEST_REMOTE_NAME
    assert remote.url == TEST_REMOTE_URL


def test_update_remotes(conan_api: ConanBaseUnifiedApi, new_remote):
    new_remote_name, _ = new_remote()

    conan_api.update_remote(new_remote_name, "http://localhost:9304", True)

    remote = conan_api.get_remote(new_remote_name)
    assert remote  # not None
    assert remote.url == "http://localhost:9304"
    assert remote.verify_ssl

    # test reorder
    conan_api.update_remote(new_remote_name, "http://localhost:9304", True, 0)
    remotes = conan_api.get_remotes()
    assert remotes[0].name == new_remote_name


def test_rename_remotes(conan_api: ConanBaseUnifiedApi, new_remote):
    new_remote_name, _ = new_remote()

    renamed_name = "new_ng_last_final"
    conan_api.rename_remote(new_remote_name, renamed_name)
    with check:
        assert "new_ng_last_final" in conan_api.get_remote_names()
    with check:
        assert new_remote_name not in conan_api.get_remote_names()

    remove_remote(renamed_name)


def test_login_remote(conan_api: ConanBaseUnifiedApi):
    conan_api.login_remote(TEST_REMOTE_NAME, "demo", "demo")

    with pytest.raises(Exception) as excinfo:
        conan_api.login_remote(TEST_REMOTE_NAME, "demo", "abc")
    




import tempfile
from shutil import copyfile
from pathlib import Path

from conan_unified_api import ConanInfoCache
from conan_unified_api.types import ConanRef as CFR
from test.conftest import PathSetup


def test_new_cache():
    """
    Test, if a new cache file is generated, if it does not exist.
    """
    temp_dir = Path(tempfile.mkdtemp())
    ConanInfoCache(temp_dir)
    assert (temp_dir / ConanInfoCache.CACHE_FILE_NAME).exists()


def test_read_cache(repo_paths: PathSetup):
    """
    Test reading from a cache file. Check internal state and use public API.
    """
    temp_cache_path = Path(tempfile.mkdtemp()) / ConanInfoCache.CACHE_FILE_NAME
    copyfile(str(repo_paths.testdata_path / "cache" / "cache_read.json"), str(temp_cache_path))

    cache = ConanInfoCache(temp_cache_path.parent)
    assert cache._remote_packages == {
        "my_package": {"user": [
            "1.0.0/channel",
            "2.0.0/channel"]
        },
        "other_package": {"others": [
            "1.0.0/testing",
            "2.0.0/stable",
            "3.0.0/stable"],
            "local": [
                "1.0.0/testing"
        ]
        }
    }
  
    pkgs = cache.get_similar_remote_pkg_refs("my_package", "user")
    assert len(pkgs) == 2
    assert CFR.loads("my_package/1.0.0@user/channel") in pkgs
    assert CFR.loads("my_package/2.0.0@user/channel") in pkgs


def test_read_and_delete_corrupt_cache(repo_paths: PathSetup):
    """Test, that an invalid jsonfile is deleted and a new one created"""
    temp_cache_path = Path(tempfile.mkdtemp()) / ConanInfoCache.CACHE_FILE_NAME
    copyfile(str(repo_paths.testdata_path / "cache" / "cache_read_corrupt.json"), str(temp_cache_path))

    ConanInfoCache(temp_cache_path.parent)
    info = ""
    with open(temp_cache_path) as tcf:
        info = tcf.read()
    assert info == ""


def test_update_cache(repo_paths: PathSetup):
    """
    Test, if updating with new values appends/updates the values correctly in the file
    """
    temp_cache_path = Path(tempfile.mkdtemp()) / ConanInfoCache.CACHE_FILE_NAME
    copyfile(str(repo_paths.testdata_path / "cache" / "cache_write.json"), str(temp_cache_path))

    cache = ConanInfoCache(temp_cache_path.parent)

    add_packages = [CFR.loads("new_pkg/1.0.0@me/stable"), CFR.loads("new_pkg/1.1.0@me/stable"),
                    CFR.loads("new_pkg/1.1.0@me/testing"), CFR.loads("new_pkg/2.0.0@me/stable")]
    cache.update_remote_package_list(add_packages)

    remote_pkgs = cache.get_similar_remote_pkg_refs("new_pkg", "me")
    assert set(remote_pkgs) == set(add_packages)
    remote_pkgs = cache.get_similar_remote_pkg_refs("my_package", "myself")
    assert len(remote_pkgs) == 3
    remote_pkgs = cache.get_similar_remote_pkg_refs("other_package", "*")
    assert len(remote_pkgs) == 4

    cache.update_remote_package_list(add_packages, invalidate=True)
    remote_pkgs = cache.get_similar_remote_pkg_refs("new_pkg", "me")
    assert set(remote_pkgs) == set(add_packages)
    remote_pkgs = cache.get_similar_remote_pkg_refs("my_package", "myself")
    assert not remote_pkgs



import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

from .base import INVALID_PATH_VALUE, conan_version
from .base.helper import str2bool
from .base.logger import Logger
from .types import (
    ConanException,
    ConanOptions,
    ConanPackageId,
    ConanPackagePath,
    ConanPkg,
    ConanPkgRef,
    ConanRef,
    ConanSettings,
    Remote,
)
from .unified_api import ConanBaseUnifiedApi, RemoteName

if TYPE_CHECKING:
    from conan_unified_api.cache.conan_cache import ConanInfoCache


class ConanUnifiedApi(ConanBaseUnifiedApi):
    """
    High level functions, which use only other ConanUnifiedApi functions are
    implemented here.
    """

    def __init__(
        self,
        init: bool = True,
        logger: Optional[logging.Logger] = None,
        mute_logging: bool = False,
    ):
        self.info_cache: ConanInfoCache

        if logger is None:
            self.logger = Logger()
        else:
            self.logger = logger
        self.mute_logging(mute_logging)

        if init:
            self.init_api()

    def mute_logging(self, mute_logging: bool) -> None:
        self.logger.disabled = mute_logging

    ### Remotes related methods

    def get_remote_names(self, include_disabled: bool = False) -> List[str]:
        return [remote.name for remote in self.get_remotes(include_disabled)]

    def get_remote(self, remote_name: str) -> Optional[Remote]:
        for remote in self.get_remotes(include_disabled=True):
            if remote.name == remote_name:
                return remote
        return None

    ### Install related methods ###

    def install_package(
        self,
        conan_ref: Union[ConanRef, str],
        package: ConanPkg,
        update: bool = True,
        remote_name: Optional[str] = None,
    ) -> Tuple[ConanPackageId, ConanPackagePath]:
        package_id = package.get("id", "")
        options = package.get("options", {})
        settings = package.get("settings", {})
        try:
            installed_id, package_path = self.install_reference(
                conan_ref,
                update=update,
                conan_settings=settings,
                conan_options=options,
                remote_name=remote_name,
            )
            if installed_id != package_id:
                Logger().warning(
                    f"Installed {installed_id} instead of selected {package_id}. This can "
                    "happen, if there transitive settings changed in comparison to build time."
                )
        except ConanException as e:
            Logger().error(f"Can't install package '{conan_ref}': {e!s}")
            installed_id = ""
            package_path = Path(INVALID_PATH_VALUE)
        return installed_id, package_path

    def get_path_with_auto_install(
        self,
        conan_ref: Union[ConanRef, str],
        conan_options: Optional[ConanOptions] = None,
        update: bool = False,
    ) -> Tuple[ConanPackageId, ConanPackagePath]:
        """Return the pkg_id and package folder of a conan reference
        and auto-install it with the best matching package, if it is not available"""
        if not update:
            pkg_id, path = self.get_best_matching_local_package_path(conan_ref, conan_options)
            if pkg_id:
                return pkg_id, path
            Logger().info(
                f"'{conan_ref}' with options {conan_options!r} is not installed. "
                "Searching for packages to install..."
            )

        pkg_id, path = self.install_best_matching_package(
            conan_ref, conan_options, update=update
        )
        return pkg_id, path

    def install_best_matching_package(
        self,
        conan_ref: Union[ConanRef, str],
        conan_options: Optional[ConanOptions] = None,
        update: bool = False,
    ) -> Tuple[ConanPackageId, ConanPackagePath]:
        packages, remote = self.find_best_matching_package_in_remotes(conan_ref, conan_options)
        if not packages:
            self.info_cache.invalidate_remote_package(conan_ref)
            return ("", Path(INVALID_PATH_VALUE))
        pkg_id, package_path = self.install_package(conan_ref, packages[0], update, remote)
        if package_path.exists():
            return pkg_id, package_path
        return "", Path(INVALID_PATH_VALUE)

    ### Local References and Packages ###

    def find_best_matching_local_package(
        self, conan_ref: Union[ConanRef, str], conan_options: Optional[ConanOptions] = None
    ) -> ConanPkg:
        """Find a package in the local cache"""
        packages = self.find_best_matching_packages(conan_ref, conan_options, remote_name=None)
        # What to if multiple ones exits? - for now simply take the first entry
        if packages:
            if len(packages) > 1:
                settings = packages[0].get("settings", {})
                pkg_id = packages[0].get("id", "")
                Logger().warning(
                    f"Multiple matching packages found for '{conan_ref}'!\n"
                    f"Choosing this: {pkg_id} ({self.build_conan_profile_name_alias(settings)})"
                )
            return packages[0]
        Logger().debug(f"No matching local packages found for {conan_ref}")
        return {"id": ""}

    def get_best_matching_local_package_path(
        self, conan_ref: Union[ConanRef, str], conan_options: Optional[ConanOptions] = None
    ) -> Tuple[ConanPackageId, ConanPackagePath]:
        """Return the pkg_id and package folder of a conan reference, if it is installed."""
        package = self.find_best_matching_local_package(conan_ref, conan_options)
        if package.get("id", ""):
            return package.get("id", ""), self.get_package_folder(
                conan_ref, package.get("id", "")
            )
        return "", Path(INVALID_PATH_VALUE)

    def get_local_pkg_from_id(self, pkg_ref: ConanPkgRef) -> ConanPkg:
        """Returns an installed pkg from reference and id"""
        package = None
        for package in self.get_local_pkgs_from_ref(pkg_ref.ref):
            if package.get("id", "") == pkg_ref.id:
                return package
        return {"id": ""}

    def get_local_pkg_from_path(
        self, conan_ref: Union[ConanRef, str], path: Path
    ) -> Optional[ConanPkg]:
        """For reverse lookup - give info from path"""
        found_package = None
        for package in self.get_local_pkgs_from_ref(conan_ref):
            if self.get_package_folder(conan_ref, package.get("id", "")) == path:
                found_package = package
                break
        return found_package

    ### Remote References and Packages ###

    def get_remote_pkg_from_id(self, pkg_ref: ConanPkgRef) -> ConanPkg:
        """Returns a remote pkg from reference and id"""
        package = None
        for remote in self.get_remotes():
            packages = self.get_remote_pkgs_from_ref(pkg_ref.ref, remote.name)
            for package in packages:
                if package.get("id", "") == pkg_ref.id:
                    return package
        return {"id": ""}

    def find_best_matching_package_in_remotes(
        self, conan_ref: Union[ConanRef, str], conan_options: Optional[ConanOptions] = None
    ) -> Tuple[List[ConanPkg], RemoteName]:
        """Find a package with options in the remotes"""
        for remote in self.get_remotes():
            packages = self.find_best_matching_packages(conan_ref, conan_options, remote.name)
            if packages:
                return (packages, remote.name)
        Logger().info(
            f"Can't find a package '{conan_ref}' with options {conan_options} in the remotes"
        )
        return ([], "")

    def find_best_matching_packages(
        self,
        conan_ref: Union[ConanRef, str],
        conan_options: Optional[ConanOptions] = None,
        remote_name: Optional[str] = None,
    ) -> List[ConanPkg]:
        """
        This method tries to find the best matching packages either locally or in a remote,
        based on the users machine and the supplied options.
        """
        if conan_options is None:
            conan_options = {}

        found_pkgs: List[ConanPkg] = []
        default_settings: ConanSettings = {}
        try:
            # dynamic prop is ok in try-catch
            default_settings = self.get_default_settings()
            query = (
                f"(arch=None OR arch={default_settings.get('arch')})"
                f" AND (os=None OR os={default_settings.get('os')})"
            )
            if conan_version.major == 1:
                query += (
                    f" AND (arch_build=None OR arch_build={default_settings.get('arch_build')})"
                    f" AND (os_build=None OR os_build={default_settings.get('os_build')})"
                )
            found_pkgs = self.get_remote_pkgs_from_ref(conan_ref, remote_name, query)
        except Exception:  # no problem, next
            return []
        if not found_pkgs:
            return []

        # remove debug releases
        no_debug_pkgs = list(
            filter(
                lambda pkg: pkg.get("settings", {}).get("build_type", "").lower() != "debug",
                found_pkgs,
            )
        )
        # check, if a package remained and only then take the result
        if no_debug_pkgs:
            found_pkgs = no_debug_pkgs

        # filter the found packages by the user options
        if conan_options:
            for pkg in found_pkgs:
                self._convert_options_to_native_ref_values(
                    pkg.get("options", {}), conan_options
                )

            # found_pkgs = list(filter(lambda pkg:
            #                          self._are_option_compatible(
            #                              conan_options, pkg.get("options", {})),
            #             found_pkgs))
            found_pkgs = list(
                filter(
                    lambda pkg: conan_options.items() <= pkg.get("options", {}).items(),
                    found_pkgs,
                )
            )

            if not found_pkgs:
                return found_pkgs
        # get a set of existing options and reduce default options with them
        min_opts_set = {frozenset(tuple(pkg.get("options", {}).keys())) for pkg in found_pkgs}
        min_opts_list = frozenset()
        if min_opts_set:
            min_opts_list = min_opts_set.pop()

        # this calls external code of the recipe
        _, default_options = self.get_options_with_default_values(conan_ref, remote_name)

        if default_options:
            default_options = dict(
                filter(lambda opt: opt[0] in min_opts_list, default_options.items())
            )
            # patch user input into default options to combine the two
            default_options.update(conan_options)
            # convert vals to string TODO use new fcn
            default_str_options: Dict[str, str] = dict(
                [key, str(value)] for key, value in default_options.items()
            )
            if len(found_pkgs) > 1:
                comb_opts_pkgs = list(
                    filter(
                        lambda pkg: default_str_options.items()
                        <= pkg.get("options", {}).items(),
                        found_pkgs,
                    )
                )
                if comb_opts_pkgs:
                    found_pkgs = comb_opts_pkgs

        # now we have all matching packages, but with potentially different compilers
        # reduce with default settings
        if len(found_pkgs) > 1:
            same_comp_pkgs = list(
                filter(
                    lambda pkg: default_settings.get("compiler", "")
                    == pkg.get("settings", {}).get("compiler", ""),
                    found_pkgs,
                )
            )
            if same_comp_pkgs:
                found_pkgs = same_comp_pkgs

            same_comp_version_pkgs = list(
                filter(
                    lambda pkg: default_settings.get("compiler.version", "")
                    == pkg.get("settings", {}).get("compiler.version", ""),
                    found_pkgs,
                )
            )
            if same_comp_version_pkgs:
                found_pkgs = same_comp_version_pkgs
        return found_pkgs

    ### Helper methods ###

    @staticmethod
    def _resolve_default_options(default_options_raw: Any) -> ConanOptions:
        """Default options can be a a dict or name=value as string, or a tuple of it"""
        default_options: Dict[str, Any] = {}
        if default_options_raw and isinstance(default_options_raw, str):
            default_option_str = default_options_raw.split("=")
            default_options.update({default_option_str[0]: default_option_str[1]})
        elif default_options_raw and isinstance(default_options_raw, (list, tuple)):
            for default_option in default_options_raw:
                default_option_str = default_option.split("=")
                default_options.update({default_option_str[0]: default_option_str[1]})
        else:
            default_options = default_options_raw
        ConanUnifiedApi._convert_options_to_str_values(default_options)
        return default_options

    @staticmethod
    def generate_canonical_ref(conan_ref: Union[ConanRef, str]) -> str:
        conan_ref = ConanUnifiedApi.conan_ref_from_reflike(conan_ref)
        if conan_ref.user is None and conan_ref.channel is None:
            return str(conan_ref) + "@_/_"
        return str(conan_ref)

    @staticmethod
    def conan_ref_from_reflike(conan_ref: Union[ConanRef, str]) -> ConanRef:
        if isinstance(conan_ref, str):
            return ConanRef.loads(conan_ref)
        return conan_ref

    @staticmethod
    def build_conan_profile_name_alias(conan_settings: ConanSettings) -> str:
        if not conan_settings:
            return "No Settings"

        os = conan_settings.get("os", "")
        if not os:
            os = conan_settings.get("os_target", "")
            if not os:
                os = conan_settings.get("os_build", "")

        arch = conan_settings.get("arch", "")
        if not arch:
            arch = conan_settings.get("arch_target", "")
            if not arch:
                arch = conan_settings.get("arch_build", "")
        if arch == "x86_64":  # shorten x64
            arch = "x64"

        comp = conan_settings.get("compiler", "")
        if comp == "Visual Studio":
            comp = "vs"
        comp_ver = conan_settings.get("compiler.version", "")
        comp_text = comp.lower() + comp_ver.lower()

        comp_toolset = conan_settings.get("compiler.toolset", "")

        bt = conan_settings.get("build_type", "")

        alias = os
        for item in [arch.lower(), comp_text, comp_toolset.lower(), bt.lower()]:
            if item:
                alias += "_" + item

        return alias

    def get_remotes_from_same_server(self, remote: Remote) -> List[Remote]:
        """
        Pass in a remote and return all other remotes with the same base url.
        Currently only for artifactory links.
        """
        remote_groups = self._get_remote_groups()
        for remotes in remote_groups.values():
            for check_remote in remotes:
                if check_remote == remote:
                    return remotes
        return [remote]

    def _get_remote_groups(self) -> Dict[str, List[Remote]]:
        """
        Try to group similar URLs (currently only for artifactory links)
        and return them in a dict grouped by the full URL.
        """
        remote_groups: Dict[str, List[Remote]] = {}
        for remote in self.get_remotes(include_disabled=True):
            if "artifactory" in remote.url:
                # try to determine root address
                possible_base_url = "/".join(remote.url.split("/")[0:3])
                if not remote_groups.get(possible_base_url):
                    remote_groups[possible_base_url] = [remote]
                else:
                    remotes = remote_groups[possible_base_url]
                    remotes.append(remote)
                    remote_groups.update({possible_base_url: remotes})
            else:
                remote_groups[remote.url] = [remote]
        return remote_groups

    @staticmethod
    def _convert_options_to_str_values(options: ConanOptions) -> Dict[str, str]:
        """Convert "ANY" to ["ANY"]: This is done to ensure compatiblity between
        Conan 1 (accepts both) and 2 (accepts only list)"""
        for key, value in options.items():
            if value == "ANY":
                options[key] = ["ANY"]
                continue
            if isinstance(value, list):
                options[key] = list(map(str, value))
                continue
            options[key] = str(value)
        return options

    @staticmethod
    def _convert_options_to_native_ref_values(
        options_to_convert: ConanOptions,
        ref_options: ConanOptions,
    ) -> None:
        for key, value in options_to_convert.items():
            if str(value) not in ["True", "False", "0"] or key not in ref_options:
                continue
            if isinstance(ref_options[key], bool):
                options_to_convert[key] = str2bool(value)
            elif isinstance(ref_options[key], str):
                options_to_convert[key] = str(value)

    @staticmethod
    def _are_option_compatible(options_ref: ConanOptions, options_other: ConanOptions) -> bool:
        options_ref_str = ConanUnifiedApi._convert_options_to_str_values(options_ref.copy())
        options_other_str = ConanUnifiedApi._convert_options_to_str_values(options_other.copy())
        return options_ref_str.items() <= options_other_str.items()



import os

import pytest
from conan_unified_api.types import ConanRef
from test import TEST_REF
from pytest_check import check

from conan_unified_api import conan_version
from conan_unified_api.unified_api import ConanBaseUnifiedApi
from test.conan_helper import add_editable, remove_editable

TEST_EDITABLE_REF = "example/9.9.9@local/editable"


@pytest.fixture
def editable_path(repo_paths):
    """ An editable target path needs a conanfile.py in it, return the original conanfile path in testdata"""
    path = repo_paths.testdata_path / "conan"
    if conan_version.major == 1:
        path /= "conanfile.py"
    elif conan_version.major == 2:
        path /= "conanfileV2.py"
    return path


@pytest.fixture
def new_editable():
    """ Fixture factory for multiple new editables. Cleans up after each testcase. """
    editable_refs = []

    def _add_editable(ref, path, output_path=None):
        remove_editable(ref) # if it exists with another path
        add_editable(ref, path, output_path)
        editable_refs.append(ref)
        return

    yield _add_editable

    for ref in editable_refs:
        remove_editable(ref)


def test_get_editable(conan_api: ConanBaseUnifiedApi, editable_path, new_editable):
    broken_editable = conan_api.get_editable(TEST_REF)
    with check:
     assert broken_editable is None
    
    new_editable(TEST_EDITABLE_REF, editable_path)
    editable = conan_api.get_editable(TEST_EDITABLE_REF)
    assert editable
    assert editable.conan_ref == TEST_EDITABLE_REF
    assert editable.path == str(editable_path)


def test_get_editables_package_path(conan_api: ConanBaseUnifiedApi, editable_path, new_editable):
    new_editable(TEST_EDITABLE_REF, editable_path)

    assert conan_api.get_editables_package_path(TEST_EDITABLE_REF) == editable_path


def test_get_editables_output_folder(conan_api: ConanBaseUnifiedApi, editable_path, 
                                     new_editable, repo_paths):
    new_editable(TEST_EDITABLE_REF, editable_path, repo_paths.testdata_path)

    assert conan_api.get_editables_output_folder(
        TEST_EDITABLE_REF) == repo_paths.testdata_path


def test_get_editable_references(conan_api: ConanBaseUnifiedApi, new_editable, editable_path):
    new_editable(TEST_EDITABLE_REF, editable_path)
    new_editable(TEST_EDITABLE_REF + "_2", editable_path)

    refs = conan_api.get_editable_references()
    assert ConanRef.loads(TEST_EDITABLE_REF) in refs
    assert ConanRef.loads(TEST_EDITABLE_REF + "_2") in refs


def test_add_remove_editable(conan_api: ConanBaseUnifiedApi, editable_path, repo_paths):
    os.system(f"conan editable remove {TEST_EDITABLE_REF}")

    conan_api.add_editable(TEST_EDITABLE_REF, editable_path, repo_paths.testdata_path)
    editable = conan_api.get_editable(TEST_EDITABLE_REF)
    assert editable
    assert editable.conan_ref == TEST_EDITABLE_REF
    assert editable.path == str(editable_path)
    assert editable.output_folder == str(repo_paths.testdata_path)

    conan_api.remove_editable(TEST_EDITABLE_REF)
    assert conan_api.get_editable(TEST_EDITABLE_REF) is None
